#===============
# SETUP
#===============

from helipad import Helipad, MultiLevel, Agent
from math import floor
from random import uniform, choice

heli = Helipad()
heli.name = 'Illicit Supply Chains'
heli.agents.order = 'random'

heli.agents.addPrimitive('firm', MultiLevel, dflt=20, priority=1)
heli.agents.removePrimitive('agent')
mapPlot = heli.spatial(dim=(5,1), wrap=False)
mapPlot.config({
	'patchProperty': 'x',
	'patchAspect': 2
})
heli.params['x'].type = 'hidden'
heli.params['y'].type = 'hidden'

#Initialize submodels
@heli.hook
def firmInit(firm, model):
	firm.param('num_agent', 20)

def firmsize(self): return len(self.agents['agent'])*5
MultiLevel.size = property(firmsize)
mapPlot.config('agentSize', 'size')

#===============
# BEHAVIOR
#===============

heli.param('num_firm', 20)

@heli.hook
def firmStep(firm, model, stage):
	if len(firm.agents['agent']):
		choice(firm.agents['agent']).transfer(choice(list(firm.patch.agentsOn))) #One agent moves to a random firm on the same patch

@heli.hook
def modelPostSetup(model):
	for n,a in enumerate(model.agents['firm']):
		stage = floor(n/4)
		a.moveTo(uniform(stage-0.4, stage+0.4), uniform(-0.5, 0.5))
	
	#Create links after all firms have been assigned to a stage
	for f in model.agents['firm']:
		if f.patch.x > 0: f.edges.add(choice(list(f.patch.left.agentsOn)))


#===============
# LAUNCH THE GUI
#===============

heli.launchCpanel()


# A group-selection model of the evolution of cooperation with deme extinction
# Loosely based on chapter 7.1 of Bowles and Gintis, "A Cooperative Species"

#===============
# SETUP
#===============

from random import choice
from math import exp
import numpy.random as nprand
from helipad import Helipad, MultiLevel

heli = Helipad()
heli.name = 'Deme Selection'
heli.agents.order = ['linear', 'linear', 'match']
heli.stages = 3 #Stage 1 for intra-demic competition, stage 2 for reproduction, stage 3 for war

heli.agents.addPrimitive('deme', MultiLevel, dflt=20, priority=1)
heli.agents.removePrimitive('agent')
heli.params.add('b', 'Benefit conferred', 'slider', dflt=6, opts={'low': 0, 'high': 10, 'step': 1})
heli.params.add('c', 'Cost incurred', 'slider', dflt=2, opts={'low': 0, 'high': 10, 'step': 1})
heli.params.add('k', 'Likelihood of war', 'slider', dflt=0.25, opts={'low': 0, 'high': 1, 'step': 0.01})

#===============
# BEHAVIOR
#===============

def strength(self):
	return sum(a.stocks['payoff'] for a in self.agents['agent'])
MultiLevel.strength = property(strength)

def agentStep(agent, deme, stage):
	if agent.breed == 'altruist':
		beneficiary = choice(deme.agents['agent'])
		agent.stocks['payoff'] -= deme.model.param('c')
		beneficiary.stocks['payoff'] += deme.model.param('b')

#deme is the MultiLevel object, which inherits from both baseAgent and Helipad.
@heli.hook
def demeInit(deme, model):
	deme.param('num_agent', 20)
	deme.agents.addBreed('altruist', '#009900')
	deme.agents.addBreed('selfish', '#990000')
	deme.goods.add('payoff', '#000099', 1)
	deme.hooks.add('agentStep', agentStep)

#Odds of winning follow a sigmoid distribution in the difference in strength
@heli.hook('demeMatch')
def war(demes, primitive, model, stage):
	k = model.param('k')
	if not nprand.choice(2, 1, p=[1-k, k])[0]: return

	#Fight
	diff = (demes[0].strength - demes[1].strength)/10
	prob = 1/(1+exp(-diff))
	result = nprand.choice(2, 1, p=[prob, 1-prob])[0]

	#Colonize
	pop = len(demes[not result].agents['agent'])
	for a in demes[not result].agents['agent']: a.die()
	for i in range(pop):
		choice(demes[result].agents['agent']).reproduce().transfer(demes[not result])

@heli.hook
def demeStep(deme, model, stage):
	if stage > 1: deme.cutStep()

	#Reproduce
	if stage==2:
		for a in deme.agents['agent']:
			for i in range(a.stocks['payoff']): a.reproduce()
			a.die()

#Normalize population at the beginning of stage 3
@heli.hook
def modelStep(model, stage):
	if stage==3:
		targetDpop = 20
		for d in model.agents['deme']:
			while len(d.agents['agent']) > targetDpop: choice(d.agents['agent']).die()
			while len(d.agents['agent']) < targetDpop: choice(d.agents['agent']).reproduce()

#===============
# DATA AND VISUALIZATION
#===============

from helipad.visualize import TimeSeries
viz = heli.useVisual(TimeSeries)

@heli.reporter
def population(model): return sum(len(d.agents['agent']) for d in model.agents['deme'])
@heli.reporter
def altruists(model): return sum(len(d.agents['agent']['altruist']) for d in model.agents['deme'])/population(model)
@heli.reporter
def selfish(model): return sum(len(d.agents['agent']['selfish']) for d in model.agents['deme'])/population(model)
@heli.reporter
def fitness(model): return sum(sum(a.stocks['payoff'] for a in d.agents['agent']) for d in model.agents['deme'])/population(model)

viz.addPlot('pop', 'Population', selected=False)
viz.addPlot('pheno', 'Phenotypes', stack=True)
viz.addPlot('fitness', 'Fitness')
viz['pop'].addSeries('population', 'Total population', '#000000')
viz['pheno'].addSeries('altruists', 'Altruists', '#33CC33')
viz['pheno'].addSeries('selfish', 'Selfish', '#BB2222')
viz['fitness'].addSeries('fitness', 'Average Fitness', '#000099')

#===============
# LAUNCH THE GUI
#===============

heli.launchCpanel()


"""
The main model module. Import and instantiate the `Helipad` class to set up and launch a model. See https://helipad.dev for API documentation.
"""

import os, sys, warnings, asyncio, time
import gettext
from random import shuffle, choice
#from memory_profiler import profile

from helipad.visualize import BaseVisualization, TimeSeries
from helipad.helpers import *
from helipad.param import Params, Shocks
from helipad.data import Data
from helipad.agent import *

class Helipad:
	"""The main model object. https://helipad.dev/functions/model/"""
	runInit = True #for multiple inheritance. Has to be a static property

	def __init__(self, locale: str='en'):
		#Have to do this first so that i18n is available early.
		#Put it in an obscure variable and then use helpers.ï() so we don't conflict with the REPL console, which overwrites _.
		if not hasattr(self, 'breed'):
			import builtins
			builtins.__dict__['helipad_gettext'] = gettext.translation('helipad', localedir=os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))+'/locales', languages=[locale]).gettext

		#Containers
		self.agents = Agents(self)
		self.data = Data(self)
		self.params = Params(self)
		self.shocks = Shocks(self)
		self.events = Events()
		self.hooks = Hooks()
		self.goods = Goods(self)

		self.name = ''
		self.patches = []
		self.stages = 1
		self.hasModel = False		#Have we initialized?
		self.timer = False
		self.visual = None
		self.cpanel = None

		self.t = None
		self.running = False
		self._cut = False

		#Default parameters
		self.agents.addPrimitive('agent', Agent, dflt=50, low=1, high=100)

		#Decorators
		def repdec(name, fn, kwargs): self.data.addReporter(name, fn, **kwargs)
		def hookdec(name, fn, kwargs): self.hooks.add(name, fn, **kwargs)
		def buttondec(name, fn, kwargs): self.addButton(name, fn, **kwargs)
		def eventdec(name, fn, kwargs): self.events.add(name, fn, **kwargs)
		self.reporter = self.genDecorator(repdec)
		self.hook = self.genDecorator(hookdec)
		self.button = self.genDecorator(buttondec)
		self.event = self.genDecorator(eventdec)

		#A few things that only make sense to do if it's the topmost model
		if not hasattr(self, 'breed'):

			#Privileged parameters
			#Toggle the progress bar between determinate and indeterminate when stopafter gets changed
			def switchPbar(model, name, val):
				if not model.hasModel or not model.cpanel or not getattr(model.cpanel, 'progress', False): return
				if not val: model.cpanel.progress.determinate(False)
				else:
					model.cpanel.progress.determinate(True)
					model.cpanel.progress.update(model.t/val)
			self.params.add('stopafter', ï('Stop on period'), 'checkentry', False, runtime=True, config=True, entryType='int', callback=switchPbar)
			self.params.add('csv', ï('CSV?'), 'checkentry', False, runtime=True, config=True)
			self.params.add('refresh', ï('Refresh Every __ Periods'), 'slider', 20, opts=[1, 2, 5, 10, 20, 50, 100, 200, 500, 1000], runtime=True, config=True)
			self.params['shocks'] = self.shocks

			#Check for updates
			from helipad.__init__ import __version__
			import xmlrpc.client, ssl

			#Is v1 bigger than v2?
			#There's a `packaging` function to do this, but let's not bloat our dependencies.
			def vcompare(v1, v2):
				(v1, v2) = (v1.split('.'), (v2.split('.')))

				#Pad major releases with zeroes to make comparable
				maxl = max([len(v1), len(v2)])
				for v in [v1,v2]:
					if len(v)<maxl: v += [0 for i in range(maxl-len(v))]

				for k,i in enumerate(v1):
					if i > v2[k]: return True
					elif i < v2[k]: return False
				return False

			try:
				pypi = xmlrpc.client.ServerProxy('https://pypi.org/pypi', context=ssl._create_unverified_context())
				available = pypi.package_releases('helipad')
				if vcompare(available[0], __version__):
					print(ï('A Helipad update is available! Use `pip install -U helipad` to upgrade to version {}.').format(available[0]))
			except: pass #Fail silently if we're not online

	def __repr__(self):
		if self.name: return f'<Helipad: {self.name}>'
		else: return '<Helipad object>'

	def addButton(self, text: str, func, desc=None):
		"""Add a button to the control panel, in the shocks section, that runs `func` when pressed. This method is aliased by the `@model.button` function decorator, which is preferred. https://helipad.dev/functions/model/addbutton/"""
		self.shocks.add(text, None, func, 'button', True, desc)

	def param(self, param, val=None):
		"""Get or set a model parameter, depending on whether there is one or two arguments. https://helipad.dev/functions/model/param/"""
		if isinstance(param, tuple):
			#Deprecated in Helipad 1.6.1, remove in 1.8
			if len(param) > 1 and param[1] in ['breed', 'good']:
				warnings.warn(ï('Three-item parameter tuple identifiers have been deprecated. The second item can be removed.'), FutureWarning, 2)
				item = param[2]
			else: item = param[1] if len(param)>1 else None
		else: item = None
		param = self.params[param[0]] if isinstance(param, tuple) else self.params[param]

		if val is not None: param.set(val, item)
		else: return param.get(item)

	def doHooks(self, place: str, args: list):
		"""Execute registered hooks at various places in the model and return the value of the last function in the list. https://helipad.dev/functions/model/dohooks/"""
		#Take a list of hooks; go until we get a response
		if isinstance(place, list):
			for f in place:
				r = self.doHooks(f, args)
				if r is not None: return r
			return None

		if place not in self.hooks: return None
		for f in self.hooks[place]: r = f(*args)
		return r

	def useVisual(self, viz: BaseVisualization):
		"""Register a visualization class for live model visualization. Visualization classes can be imported from `helipad.visualize`, or custom visualization classes can be subclassed from `BaseVisualization`. The visualization can then be launched later using model.launchVisual(). https://helipad.dev/functions/model/usevisual/"""
		if hasattr(self, 'breed'):
			warnings.warn(ï('Visualizations can only be registered on the top-level model.'), None, 2)
			return #Doesn't matter if it's not the top-level model

		if viz is not None and not issubclass(viz, BaseVisualization):
			raise RuntimeError(ï('Visualization class must inherit from BaseVisualization.'))

		self.visual = viz(self) if viz is not None else None
		return self.visual

	def setup(self):
		"""Gather the control panel settings and initialize the model. This function runs when the "New Model" button is pressed, and should not be called by user code. https://helipad.dev/functions/model/setup/"""
		if self.hasModel: self.terminate()
		self.doHooks('modelPreSetup', [self])
		self.t = 0

		#Blank breeds for any primitives not otherwise specified
		for p in self.agents.values():
			if not p.breeds: p.breeds.add('', '#000000')

		#SERIES AND REPORTERS
		#Breeds and goods should already be registered at this point

		self.data.reset()
		for e in self.events.values(): e.reset()
		defPrim = 'agent' if 'agent' in self.agents else next(iter(self.agents))

		def pReporter(param, item=None):
			def reporter(model): return param.get(item)
			return reporter

		#Add reporters for parameters
		for n,p in self.params.items():
			if p.type == 'hidden' or getattr(p, 'config', False): continue	#Skip hidden and config parameters
			if p.per is None: self.data.addReporter(n, pReporter(p))
			elif p.per == 'good':
				for good in p.pKeys:
					self.data.addReporter(n+'-'+good, pReporter(p, good))
			elif p.per == 'breed':
				for breed in p.pKeys:
					self.data.addReporter(p.prim+'-'+n+'-'+breed, pReporter(p, breed))

		if self.goods.money is not None:
			self.data.addReporter('M0', self.data.agentReporter('stocks', 'all', good=self.goods.money, stat='sum'))
			if isinstance(self.visual, TimeSeries) and 'money' in self.visual:
				self.visual['money'].addSeries('M0', ï('Monetary Base'), self.goods[self.goods.money].color)

		#Unconditional variables to report
		# self.data.addReporter('utility', self.data.agentReporter('utils', defPrim))

		#Per-breed and per-good series and reporters
		#Don't put lambda functions in here, or the variable pairs will be reported the same, for some reason.
		for breed, b in self.agents[defPrim].breeds.items():
			self.data.addReporter('utility-'+breed, self.data.agentReporter('utils', defPrim, breed=breed))
			if isinstance(self.visual, TimeSeries):
				self.visual['utility'].addSeries('utility-'+breed, breed.title()+' '+ï('Utility'), b.color)

		if len(self.goods) >= 2:
			for good, g in self.goods.nonmonetary.items():
				self.data.addReporter('demand-'+good, self.data.agentReporter('currentDemand', 'all', good=good, stat='sum'))
				if isinstance(self.visual, TimeSeries) and 'demand' in self.visual:
					self.visual['demand'].addSeries('demand-'+good, good.title()+' '+ï('Demand'), g.color)

		#Initialize agents
		for prim, ags in self.agents.items():
			ags.clear()																	#Clear any surviving agents from last run
			self.agents.initialize(self.param('num_'+prim), prim, self, force=True)		#Force is so we can call initialize() before instantiating hasModel

		#Start progress bar
		#Put this here and not in .start() because it'll flash on unpause otherwise
		if self.cpanel:
			st = self.param('stopafter')
			self.cpanel.progress.determinate(st and isinstance(st, int))

		for param in self.params.values():
			if not param.runtime: param.disable() #Disable parameters that can't be changed during runtime

		#Patch our async functions for compatibility with Spyder's event loop
		if isIpy() and not isNotebook():
			try:
				import nest_asyncio
				nest_asyncio.apply()
			except:
				raise ImportError(ï('nest_asyncio is required to run Helipad from Spyder.'))

		self.hasModel = True
		self.doHooks('modelPostSetup', [self])

	def cutStep(self):
		"""When called from inside an `agentStep` or `match` hook, skip stepping the rest of the agents that stage and proceed to the next (or to the next period in single-stage models). https://helipad.dev/functions/model/cutstep/"""
		self._cut = True

	def step(self, stage: int=1):
		"""Step the model, i.e. run through the `step()` functions of all the agents and increment the timer by one. This method is called automatically while the model is running, and should not generally be called in user code. https://helipad.dev/functions/model/step/"""
		self.t += 1
		self.doHooks('modelPreStep', [self])

		#Reset per-period variables
		#Have to do this all at once at the beginning of the period, not when each agent steps
		for p in self.agents.values():
			for a in p: a.currentDemand = {g:0 for g in self.goods}

		self.shocks.step()

		def sortFunc(model: Helipad, stage: int, func):
			def sf(agent): return func(agent, model, stage)
			return sf

		for self.stage in range(1, self.stages+1):
			self._cut = False
			self.doHooks('modelStep', [self, self.stage])

			#Sort agents and step them
			for prim, lst in self.agents.items():
				order = lst.order or self.agents.order
				if isinstance(order, list): order = order[self.stage-1]
				if order == 'random': shuffle(lst)

				#Can't do our regular doHooks() here since we want to pass the function to .sort()
				#From the user's perspective though, this doesn't matter
				#Do the more specific sorts last
				ordhooks = (self.hooks['order'] if 'order' in self.hooks else []) + (self.hooks[prim+'Order'] if prim+'Order' in self.hooks else [])
				for o in ordhooks: lst.sort(key=sortFunc(self, self.stage, o))

				#Copy the agent list to keep it constant for a given loop because modifying it
				#while looping (e.g. if an agent dies or reproduces) will screw up the looping
				agentpool = list(lst)

				#Matching model
				if 'match' in order:
					matchN = int(mn[1]) if len(mn := order.split('-')) > 1 else 2
					matchpool = list(lst)
					while len(matchpool) > len(agentpool) % matchN and not self._cut:
						agents = []
						for a in range(matchN):
							agent = choice(matchpool)
							matchpool.remove(agent)
							agents.append(agent)

							#Allow selection of matches, but only on the first agent
							if a==0:
								others = self.doHooks('matchSelect', [agent, matchpool, self, self.stage])
								if others is not None:
									if (isinstance(others, baseAgent) and matchN==2): others = [others]
									if (isinstance(others, list) and len(others)==matchN-1):
										for other in others: matchpool.remove(other)
										agents += others
										break
									else: raise ValueError(ï('matchSelect did not return the correct number of agents.'))

						accept = self.doHooks('matchAccept', agents)
						if accept is not None and not accept:
							matchpool = agents + matchpool #Prepend agents to decrease the likelihood of a repeat matching
							continue
						for a in agents: agent.step(self.stage)
						self.doHooks([prim+'Match', 'match'], [agents, prim, self, self.stage])

					#Step any remainder agents
					for agent in matchpool:
						if not self._cut: agent.step(self.stage)

				#Activation model
				else:
					for a in agentpool:
						if not self._cut: a.step(self.stage)

		self.data.collect(self)
		for e in self.events.values():
			if (not e.triggered or e.repeat) and e.check(self) and self.visual is not None and not self.visual.isNull and e.name!=self.param('stopafter'):
				self.visual.event(self.t, **e.kwargs)
		self.doHooks('modelPostStep', [self])
		return self.t

	#This is split out as an async function to allow user input while running the loop in Jupyter
	async def run(self):
		if self.timer:
			begin = time.time()
		while self.running:
			t = self.step()
			st = self.param('stopafter')

			if t%self.param('refresh')==0:
				if self.timer: t2 = time.time()
				if self.cpanel and st and isinstance(st, int): self.cpanel.progress.update(t/st)

				#Update visualizations
				if self.visual is not None and not self.visual.isNull:
					await asyncio.sleep(0.001) #Listen for keyboard input
					data = self.data.getLast(t - self.visual.lastUpdate)

					self.visual.refresh(data)
					self.visual.lastUpdate = t

					self.doHooks('visualRefresh', [self, self.visual])

				elif self.cpanel:
					if isNotebook(): await asyncio.sleep(0.001) #Listen for keyboard input
					else: self.cpanel.update() #Make sure we don't hang the interface if plotless

				# Performance indicator
				if self.timer:
					newtime = time.time()
					output = ï('Period {0}: {1} periods/second ({2}% model, {3}% visuals)').format(t, round(self.param('refresh')/(newtime-begin),2), round((t2-begin)/(newtime-begin)*100,2), round((newtime-t2)/(newtime-begin)*100,2))
					if isBuffered():
						if isNotebook(): os.write(1, bytes('\r'+output, 'utf-8'))
						else: sys.stdout.write('\r'+output)
					else: print(output)
					begin = newtime

			if st:
				stop = self.events[st].triggered if isinstance(st, str) else t>=st
				if stop: self.terminate()
		if self.timer and isBuffered(): print('\r') #Newline on pause so the output doesn't interfere with the console

	#The *args allows it to be used as an Ipywidgets callback
	#@profile #To test memory usage
	def start(self, *args):
		"""Start a model, running `model.setup()` if `model.hasModel` is `False`, and otherwise resuming the existing model. This function is called by `model.launchVisual()`, but does not call it. Use that function to start the model instead if output plots are desired. https://helipad.dev/functions/model/start/"""
		#Start the progress bar
		if self.cpanel:
			self.cpanel.progress.start()
			self.cpanel.runButton.run()

		self.doHooks('modelStart', [self, self.hasModel])
		if not self.hasModel: self.setup()
		self.running = True

		if isNotebook(): asyncio.ensure_future(self.run()) #If Jupyter, it already has an event loop
		else: asyncio.run(self.run())	#If Tkinter, it needs an event loop

	def stop(self, *args):
		"""Pause the model, allowing it to be subsequently resumed. https://helipad.dev/functions/model/stop/"""
		self.running = False
		if self.cpanel:
			self.cpanel.progress.stop()
			self.cpanel.runButton.pause()
		self.doHooks('modelStop', [self])

	def terminate(self, evt=False):
		"""Terminate the running model and write the data to disk, if applicable. https://helipad.dev/functions/model/terminate/"""
		if not self.hasModel: return
		self.running = False
		self.hasModel = False
		if self.visual is not None: self.visual.terminate(self) #Clean up visualizations

		remainder = int(self.t % self.param('refresh')) #For some reason this returns a float sometimes?
		if remainder > 0 and self.visual is not None and not self.visual.isNull: self.visual.update(self.data.getLast(remainder)) #Last update at the end

		if self.param('csv'): self.data.saveCSV(self.param('csv'))
		if self.cpanel:
			self.cpanel.progress.done()
			self.cpanel.runButton.terminate()

		#Re-enable parameters
		for param in self.params.values():
			if param.type=='checkentry' and param.event: continue
			if not param.runtime: param.enable()

		self.doHooks('terminate', [self, self.data.dataframe])

	#param is a string (for a global param), a name,object,item,primitive tuple (for per-breed or per-good params), or a list of such
	def paramSweep(self, param, reporters=None):
		"""Repeatedly run the model while systematically varying one or more parameter values. Possible values to be swept are specified when the parameter is registered. https://helipad.dev/functions/model/paramsweep/"""
		if not self.param('stopafter'): raise RuntimeError(ï('Can\'t do a parameter sweep without the value of the \'stopafter\' parameter set.'))

		#Standardize format and get the Param objects
		if not isinstance(param, list): param = [param]
		params = {}
		for p in param:
			if not isinstance(p, tuple): p = (p,)
			pobj = self.params[p[0]]
			params['-'.join(p)] = (p, pobj)

		#Generate the parameter space, a list of dicts
		from itertools import product
		space = [{p[0]:run[k] for k,p in enumerate(params.items())} for run in product(*[p[1].range for p in params.values()])]

		#Run the model
		alldata = []
		for i,run in enumerate(space):
			print('Run',str(i+1)+'/'+str(len(space))+':',', '.join([k+'='+('\''+v+'\'' if isinstance(v, str) else str(v)) for k,v in run.items()])+'…')
			for p in self.params.values():
				if not getattr(p, 'config', False): p.reset()

			for k,v in run.items(): params[k][1].set(v, params[k][0][2] if params[k][1].per is not None else None)
			self.setup()
			self.start()

			if reporters is not None: data = pandas.DataFrame({k:self.data.all[k] for k in reporters})
			else: data = self.data.dataframe

			events = [Item(name=e.name, triggered=e.triggered, data=e.data) for e in self.events.values()]

			alldata.append(Item(vars=run, data=data, events=events))
		return alldata

	def spatial(self, *args, **kwargs):
		from helipad.spatial import spatialSetup
		return spatialSetup(self, *args, **kwargs)

	# Only works on Mac. Also Gnureadline borks everything, so don't install that.
	# Has to be called *after* Cpanel.__init__() is called, or the cpanel object won't be available.
	def debugConsole(self):
		"""Launch a REPL console to interact with the model after launch. Requires to be run in a buffered console. `self` will refer to the model object."""
		if sys.platform=='darwin' and isBuffered():
			try:
				import code, readline # Readline doesn't look like it's doing anything here, but it enables certain console features
				env = globals().copy()
				env['self'] = self
				self.console = code.InteractiveConsole(locals=env)
				self.console.interact()
			except ModuleNotFoundError: print(ï('Error initializing the debug console. Make sure the `readline` and `code` modules are installed.'))
			except SystemExit: pass

	def launchCpanel(self, console: bool=True):
		"""Launch the control panel, either in a Tkinter or a Jupyter environment, as appropriate. https://helipad.dev/functions/model/launchcpanel/"""
		if hasattr(self, 'breed'): warnings.warn(ï('Control panel can only be launched on the top-level model.'), None, 2)

		self.doHooks('CpanelPreLaunch', [self])

		#Set our agents slider to be a multiple of how many agent types there are
		#Do this down here so we can have breeds registered before determining options
		for k,p in self.agents.items():
			if self.params['num_'+k].type != 'hidden':
				l = len(p.breeds)
				if not l: continue
				self.params['num_'+k].opts['low'] = makeDivisible(self.params['num_'+k].opts['low'], l, 'max')
				self.params['num_'+k].opts['high'] = makeDivisible(self.params['num_'+k].opts['high'], l, 'max')
				self.params['num_'+k].opts['step'] = makeDivisible(self.params['num_'+k].opts['low'], l, 'max')
				self.params['num_'+k].value = makeDivisible(self.params['num_'+k].value, l, 'max')
				self.params['num_'+k].default = makeDivisible(self.params['num_'+k].default, l, 'max')

		if isNotebook():
			from helipad.cpanelJupyter import Cpanel, SilentExit
			if self.cpanel: self.cpanel.invalidate(ï('Control panel was redrawn in another cell.'))
			self.cpanel = Cpanel(self)
			self.doHooks('CpanelPostInit', [self.cpanel])
			raise SilentExit() #Don't blow past the cpanel if doing "run all"
		else:
			from helipad.cpanelTkinter import Cpanel
			self.cpanel = Cpanel(self)
			self.doHooks('CpanelPostInit', [self.cpanel]) #Want the cpanel property to be available here, so don't put in cpanel.py
			if console: self.debugConsole()
			self.cpanel.mainloop()		#Launch the control panel

		self.doHooks('GUIClose', [self]) #This only executes after all GUI elements have closed

	def launchVisual(self):
		"""Launch the visualization window from the class registered in `model.useVisual()`, and start the model. https://helipad.dev/functions/model/launchvisual/"""
		if self.visual is None or self.visual.isNull:
			if not self.cpanel:
				print(ï('No visualizations available. To run the model with no GUI, use model.start() instead.'))
				return
			if not self.param('stopafter') or not (self.param('csv') or 'terminate' in self.hooks):
				print(ï('Running from the control panel with no visualization requires a stop condition, and either CSV export or a terminate hook.'))
				return

		self.setup()

		if self.visual is not None and not self.visual.isNull:
			self.visual.launch(ï('{}Data Plots').format(self.name+(' ' if self.name!='' else '')))
		else: #Headless
			self.params['stopafter'].disable()
			self.params['csv'].disable()

		self.doHooks('visualLaunch', [self, self.visual])

		#If we're running in cpanel-less mode, hook through a Tcl loop so it doesn't exit on pause
		if not self.cpanel and not self.visual.isNull and not isNotebook():
			from tkinter import Tcl
			root = Tcl()
			root.after(1, self.start)
			self.debugConsole()
			root.mainloop()
		else: self.start() #As long as we haven't already started

	# Generates function decorators for hooks, reporters, etc.
	def genDecorator(self, todo):
		"""Return a function which can either serve as a decorator itself or return another decorator function, for use in various decorators (`@model.hook`, `@model.reporter`, and `@model.button`). https://helipad.dev/functions/model/gendecorator/"""
		def dec(name=None, **kwargs):
			if callable(name): func, name, isDec = (name, None, True)
			else: isDec = False

			#Is a decorator
			def rep1(fn):
				namn = name #If I do anything with name here, Python throws UnboundLocalError. Possibly a Python bug?
				if namn is None: namn=fn.__name__
				todo(namn, fn, kwargs)
				return fn

			return rep1(func) if isDec else rep1
		return dec

	#
	#Deprecated in 1.6, remove in 1.8
	#

	@property
	def allagents(self):
		"""A list of all agents. This property is deprecated; use `model.agents.all` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('model.allagents', 'model.agents.all'), FutureWarning, 2)
		return self.agents.all

	@property
	def primitives(self):
		"""A list of primitive data. This property is deprecated; use `model.agents` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('model.primitives', 'model.agents'), FutureWarning, 2)
		return self.agents

	@property
	def order(self):
		"""The order in which to step agents. This property is deprecated; use `model.agents.order` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('model.order', 'model.agents.order'), FutureWarning, 2)
		return self.agents.order
	@order.setter
	def order(self, val):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('model.order', 'model.agents.order'), FutureWarning, 2)
		self.agents.order = val

	def addBreed(self, name, color, prim=None):
		"""Add a breed to a given primitive. This method is deprecated; use `model.agents.addBreed()` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('model.addBreed', 'model.agents.addBreed'), FutureWarning, 2)
		return self.agents.addBreed(name, color, prim)

	def createNetwork(self, density, kind='edge', prim=None):
		"""Create a network of a given density among agents. This method is deprecated; use `model.agents.createNetwork()` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('model.createNetwork', 'model.agents.createNetwork'), FutureWarning, 2)
		return self.agents.createNetwork(density, kind, prim)

	def network(self, kind='edge', prim=None, excludePatches=False):
		"""Return the network structure of a set of agents. This method is deprecated; use `model.agents.network()` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('model.network', 'model.agents.network'), FutureWarning, 2)
		return self.agents.network(kind, prim, excludePatches)

	@property
	def allEdges(self):
		"""A `dict` of all network edges between agents, organized by kind. This property is deprecated; use `model.agents.edges` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('model.allEdges', 'model.agents.edges'), FutureWarning, 2)
		return self.agents.edges._dict

	def agent(self, var, primitive=None):
		"""Retrieve an agent by ID or breed. This method is deprecated; use `model.agents[primitive][breed]` or `model.agents[id]` instead."""
		if isinstance(var, str):
			warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('model.agent(breed)', 'model.agents[primitive][breed]'), FutureWarning, 2)
			if primitive is None: primitive = next(iter(self.agents))
			return self.agents[primitive][var]
		else:
			warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('model.agent(id)', 'model.agents[id]'), FutureWarning, 2)
			return self.agents[var]

	def summary(self, var, prim=None, breed=None, good=False):
		"""Print summary statistics on an agent property. This method is deprecated; use `model.agents.summary()` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('model.summary()', 'model.agents.summary()'), FutureWarning, 2)
		return self.agents.summary(var, prim, breed, good)

class MultiLevel(baseAgent, Helipad):
	"""A class allowing multi-level agent-based models to be constructed, where the agents at one level are themselves full models with sub-agents. Inherits from both `baseAgent` and `Helipad`. https://helipad.dev/functions/multilevel/"""
	def __init__(self, breed, aId, parentModel):
		super().__init__(breed, aId, parentModel)
		self.setup()

#==================
# CONTAINER CLASSES
#==================

class Events(funcStore):
	"""Interface to add and store events that can trigger during a model on a certain user-defined criterion. https://helipad.dev/functions/events/"""
	def __init__(self):
		super().__init__()
		class Event:
			"""An event triggers on a certain user-defined criterion. When triggered, an event stores the data output at that time and registers on the visualizer. This class should not be instantiated directly; use Events.add() or the @model.event decorator instead. https://helipad.dev/functions/event/"""
			def __init__(self, name: str, trigger, repeat: bool=False, **kwargs):
				self.name = name
				self.trigger = trigger
				self.repeat = repeat
				self.kwargs = kwargs
				self.data = []
				self.triggered = []
				self.reset()

			def check(self, model: Helipad) -> bool:
				"""Run the user-defined event function to determine whether the event is triggered, and record the model state if so. If `Event.repeat==False`, return `False` so long as `Event.trigger` returns `False`, `True` the first time `Event.trigger` returns `True`, and `False` thereafter. If `Event.repeat==True`, return `True` whenever `Event.trigger` returns `True`. https://helipad.dev/functions/event/check/"""
				if self.triggered and not self.repeat: return False
				if (isinstance(self.trigger, int) and model.t==self.trigger) or (callable(self.trigger) and self.trigger(model)):
					data = {k: v[0] for k,v in model.data.getLast(1).items()}
					if self.repeat:
						self.data.append(data)
						self.triggered.append(model.t)
					else:
						self.data = data
						self.triggered = model.t
					return True

			def reset(self):
				"""Clear `Event.data` and set `Event.triggered` to `False`. https://helipad.dev/functions/event/reset/"""
				if self.repeat:
					self.data.clear()
					self.triggered.clear()
				else:
					self.data = None
					self.triggered = False
		self.Event = Event

	def add(self, name: str, function, **kwargs):
		"""Register an Event. When triggered, an event stores the data output at that time and registers on the visualizer. https://helipad.dev/functions/events/add/"""
		return super().add(name, self.Event(name, function, **kwargs))

class Goods(gandb):
	"""Interface to add and store goods that agents can own. Stored in `model.goods`. https://helipad.dev/functions/goods/"""
	def add(self, name: str, color, endowment=None, money: bool=False, props=None):
		"""Register a good that agents can carry or trade. Agents keep track of stocks of the good in `agent.stocks`. Quantities of a good can then be accessed with `agent.stocks[good]`, and properties of the good with a two-argument index, e.g. `agent.stocks[good, 'property']`. https://helipad.dev/functions/goods/add/"""
		if not props: props = {}
		if money:
			if self.money is not None:
				print(ï('Money good already specified as {}. Overriding…').format(self.money))
				self[self.money].money = False

			#Add the M0 plot once we have a money good, only if we haven't done it before
			elif (self.model.visual is None or self.model.visual.isNull) and hasattr(self.model.visual, 'plots'):
				try:
					if 'money' not in self.model.visual: self.model.visual.addPlot('money', ï('Money'), selected=False)
				except: pass #Can't add plot if re-drawing the cpanel

		props['quantity'] = endowment
		item = super().add('good', name, color, money=money, props=props)

		#Add demand plot once we have at least 2 goods, but only if the cpanel hasn't been drawn yet
		if len(self) == 2 and (self.model.visual is None or self.model.visual.isNull) and hasattr(self.model.visual, 'plots'):
			if 'demand' not in self.model.visual: self.model.visual.addPlot('demand', ï('Demand'), selected=False)

		return item

	@property
	def money(self):
		"""The name of the good serving as a numeraire. This property is set by the `money` parameter of `Goods.add()`. https://helipad.dev/functions/goods/#money"""
		for name,good in self.items():
			if good.money: return name
		return None

	@property
	def nonmonetary(self) -> dict:
		"""The subset of goods minus the money good. https://helipad.dev/functions/goods/#nonmonetary"""
		return {k:v for k,v in self.items() if not v.money}

class Hooks(funcStore):
	"""Interface to add and store hooks, allowing user-defined code to be inserted into the model. https://helipad.dev/functions/hooks/"""
	multi: bool = True

	def add(self, name: str, function, prioritize: bool=False):
		"""Inserts a function into designated places in the model’s logic. See the Hooks Reference (https://helipad.dev/glossary/hooks/) for a complete list of possible hooks and the function signatures necessary to use them. This method is aliased by the `@model.hook` function decorator, which is the preferred way to hook functions. https://helipad.dev/functions/hooks/add/"""
		if name not in self: self[name] = []
		if prioritize: self[name].insert(0, function)
		else: self[name].append(function)


"""
Classes to abstract the interface between model parameters and GUI elements. This module should not be imported directly; use `model.params.add()` instead.
"""

from itertools import combinations
from numpy import arange, random
from helipad.helpers import warnings, ï, isNotebook, Item, funcStore

class Param(Item):
	"""Base class defining a model parameter that can be set before or during runtime. https://helipad.dev/functions/param/"""
	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		if not hasattr(self, 'per'): self.per=None
		if not hasattr(self, 'value'): self.value = {b:self.defaultVal for b in kwargs['pKeys']} if self.per else self.defaultVal
		self.element = None
		self.reset() #Populate with default values

	def __repr__(self): return f'<{self.__class__.__name__}: {self.name}>'

	#Global generic:				int → int
	#Per-breed universal generic:	int → dict{int}
	#Per-breed specific generic:	dict{int} → dict{int}
	def reset(self):
		"""Reset the parameter to its default value. https://helipad.dev/functions/param/reset/"""
		if self.per is None: self.setSpecific(self.default)
		else:
			if isinstance(self.default, dict):
				for i in self.pKeys:
					self.setSpecific(self.default[i] if i in self.default else self.defaultVal, i)
			else:
				for i in self.pKeys: self.setSpecific(self.default, i)

	def setParent(self, val, item=None, updateGUI: bool=True):
		"""Common code for subclasses' `set()` methods."""
		if self.per is not None and item is None: raise KeyError(ï('A {} whose parameter value to set must be specified.').format(self.per))

		#Jupyter requires an explicit update for all parameter types.
		if updateGUI and isNotebook() and self.element is not None:
			if self.per is None: self.element.children[0].value = val
			else: self.elements[item].children[0].value = val

	#Don't override this one
	def set(self, val, item=None, updateGUI: bool=True):
		"""Set the value of the parameter to `val`. Do not call directly; use `model.param()` instead. https://helipad.dev/functions/param/set/"""
		if getattr(self, 'setter', False):
			val = self.setter(val, item)
			if val is not None: self.setSpecific(val, item)
		else: self.setSpecific(val, item, updateGUI)

	def setSpecific(self, val, item=None, updateGUI: bool=True):
		"""A generic set method to be overridden by subclasses."""
		self.setParent(val, item, updateGUI)
		if self.per is None: self.value = val
		else: self.value[item] = val

	#Don't override this one
	def get(self, item=None):
		"""Retrieve the value of the parameter. Do not call directly; use `model.param()` instead. https://helipad.dev/functions/param/get/"""
		if getattr(self, 'getter', False):
			v = self.getter(item)
			if v is not None: return v #Allow passing to the default getter
		return self.getSpecific(item)

	def getSpecific(self, item=None):
		"""A generic get method to be overridden by subclasses."""
		return self.value if self.per is None or item is None else self.value[item]

	def disabled(self, disable: bool):
		"""Enables or disables the parameter's GUI element. https://helipad.dev/functions/param/disabled/"""
		if self.element is None: return
		for e in ([self.element] if self.per is None else self.elements.values()):
			if isNotebook():
				for i in e.children: i.disabled = disable
			else:
				e.configure(state='disabled' if disable else 'normal')

	def disable(self):
		"""Disables the parameter's control panel GUI element so that it cannot be interacted with or changed by the user. https://helipad.dev/functions/param/disable/"""
		self.disabled(True)
	def enable(self):
		"""Enables the parameter's control panel GUI element so that it can be interacted with by the user. https://helipad.dev/functions/param/enable/"""
		self.disabled(False)

	@property
	def range(self):
		"""A list of possible values that the parameter can take."""
		return None

	#If a breed or good gets added after the parameter instantiation, we want to be able to keep up
	def addKey(self, key: str):
		"""Update a per-item parameter with a new item and assign it the default value specified at the `Param` object's instantiation. https://helipad.dev/functions/param/addkey/"""
		if self.per is None: warnings.warn(ï('Can\'t add keys to a global parameter…'), None, 2)
		elif not isinstance(self.default, dict):
			self.value[key] = self.default
		elif key in self.default:
			self.value[key] = self.default[key]	#Forgive out-of-order specification
		else: return True

	@property
	def defaultVal(self):
		"""Fallback value in case `param.default` is not specified on init."""
		return None

class MenuParam(Param):
	"""A parameter type that takes discrete string values. https://helipad.dev/functions/param/"""
	type = 'menu'

	def setParent(self, val, item=None, updateGUI=True):
		if self.per is not None and item is None: raise KeyError(ï('A {} whose parameter value to set must be specified.').format(self.per))
		if updateGUI and not isNotebook() and self.element is not None:
			sv = self.element if self.per is None else self.elements[item]
			sv.StringVar.set(self.opts[val]) #Saved the StringVar here
		else: super().setParent(val, item, updateGUI)

	@property
	def range(self): return self.opts.keys()

	#Choose first item of the list
	@property
	def defaultVal(self): return next(iter(self.opts))

class CheckParam(Param):
	"""A parameter type that takes a boolean value. https://helipad.dev/functions/param/"""
	defaultVal = False
	type='check'

	def setParent(self, val: bool, item=None, updateGUI: bool=True):
		if self.per is not None and item is None: raise KeyError(ï('A {} whose parameter value to set must be specified.').format(self.per))
		if updateGUI and not isNotebook() and self.element is not None:
			sv = self.element if self.per is None else self.elements[item]
			sv.BooleanVar.set(val) #Saved the BooleanVar here
		else: super().setParent(val, item, updateGUI)

	@property
	def range(self): return [False, True]

class SliderParam(Param):
	"""A parameter type that takes numerical values. https://helipad.dev/functions/param/"""
	type='slider'

	def setSpecific(self, val, item=None, updateGUI: bool=True):
		#If we're receiving a value from a Jupyter logslider, it's an index and not a value
		if not updateGUI and isNotebook() and isinstance(self.opts, list): val = self.opts[val]

		self.setParent(val, item, updateGUI)
		super().setSpecific(val, item, updateGUI)

		#Because the slider tkinter widget doesn't use a Var() object like the others, we have
		#to explicitly push the value to the slider if necessary
		if updateGUI and self.element is not None and not isNotebook():
			if self.per is None: self.element.set(val)
			else: self.elements[item].set(val)

	def getSpecific(self, item=None):
		v = super().getSpecific(item)
		#Have sliders with an int step value return an int
		if self.opts is not None and 'step' in self.opts and isinstance(self.opts['step'], int):
			if isinstance(v, dict): v = {k: int(val) for k,val in v.items()}
			else: v = int(v)
		return v

	@property
	def range(self) -> list:
		values = arange(self.opts['low'], self.opts['high'], self.opts['step']).tolist()
		values.append(self.opts['high']) #arange doesn't include the high
		return values

	def addKey(self, key: str):
		if super().addKey(key) is None: return
		self.value[key] = 0

	@property
	def defaultVal(self): return self.opts['low'] if isinstance(self.opts, dict) else self.opts[0]

	#Override in order to update log slider
	def setParent(self, val, item=None, updateGUI: bool=True):
		if self.per is not None and item is None: raise KeyError(ï('A {} whose parameter value to set must be specified.').format(self.per))

		if isNotebook() and self.element is not None and (self.per is None or item in self.elements):
			el = self.element if self.per is None else self.elements[item]

			#Regular slider update
			if updateGUI and isinstance(self.opts, dict): el.children[0].value = val

			#If it's a log slider in Jupyter, update the label, even if we're receiving from the GUI
			#And then push to the slider if we're not receiving from the GUI
			elif isinstance(self.opts, list):
				el.children[1].value = str(val)
				if updateGUI and val in self.opts: el.children[0].value = self.opts.index(val)

class CheckentryParam(Param):
	"""A parameter type that takes `False` if unchecked, or a string value if checked. https://helipad.dev/functions/param/"""
	defaultVal = ''
	type='checkentry'

	def __init__(self, **kwargs):
		self.bvar = True if 'per' not in kwargs or kwargs['per'] is None else {k:True for k in kwargs['pKeys']}
		self.svar = self.defaultVal if 'per' not in kwargs or kwargs['per'] is None else {k:self.defaultVal for k in kwargs['pKeys']}
		self.value = None
		self.event = False
		kwargs['entryType'] = int if 'entryType' in kwargs and kwargs['entryType']=='int' else str
		super().__init__(**kwargs)

	def getSpecific(self, item=None):
		#Global parameter
		if self.per is None:
			if self.name=='stopafter' and self.event: return self.svar
			elif self.element is not None and not isNotebook(): return self.element.get()
			else: return self.svar if self.bvar else False
		#Per-item parameter, get all
		elif item is None:
			if hasattr(self, 'elements') and not isNotebook(): return {k: v.get() for k,v in self.elements.items()}
			else: return {k: self.svar[k] if self.bvar[k] else False for k in self.pKeys}
		#Per-item parameter, get one
		else:
			if hasattr(self, 'elements') and item in self.elements and not isNotebook(): return self.elements[item].get()
			else: return self.svar[item] if self.bvar[item] else False

	def setSpecific(self, val, item=None, updateGUI: bool=True):
		self.setParent(val, item, False) #Don't update the GUI because it's a complex multivar type
		if self.per is None:

			#Manage setting stopafter to an event
			if self.name == 'stopafter':
				if val and isinstance(val, str):
					self.event = True
					if self.element is not None:
						self.disable()
						if isNotebook():
							self.element.children[1].value = 'Event: '+val
							self.element.children[0].value = True
							self.element.add_class('helipad_checkentry_func')
						else:
							self.element.entryValue.set('Event: '+val)
							self.element.checkVar.set(True)
							self.element.textbox.config(font=('Helvetica Neue', 12,'italic')) #Lucida doesn't have an italic?
					self.svar = val
					self.bvar = True
					return
				elif self.event and self.element is not None:
					self.event = False
					if isNotebook():
						if isinstance(val, bool): self.element.children[1].value = ''
						self.element.children[0].disabled = False
						self.element.remove_class('helipad_checkentry_func')
					else:
						if isinstance(val, bool): self.element.entryValue.set('')
						self.element.enable()
						self.element.textbox.config(font=('Lucida Grande', 12))

			if self.element is not None and not isNotebook(): self.element.set(val)
			elif isinstance(val, bool):
				self.bvar = val
				if isNotebook() and self.element is not None: self.element.children[1].disabled = not val
			elif isinstance(val, self.entryType):
				self.bvar = True
				self.svar = val
				if isNotebook() and self.element is not None: self.element.children[1].disabled = False
		else:
			if hasattr(self, 'elements') and not isNotebook(): self.elements[item].set(val)
			elif isinstance(val, bool): self.bvar[item] = val
			elif isinstance(val, self.entryType):
				self.bvar[item] = True
				self.svar[item] = val

		if updateGUI and isNotebook() and self.element is not None:
			els = self.element.children if self.per is None else self.elements[item].children
			els[0].value = val is not False
			if not isinstance(val, bool): els[1].value = str(val) #Ipy text widget requires a string

	#Only re-enable the textbox if the checkbox is checked
	def disabled(self, disable):
		if self.element is not None and isNotebook() and not disable:
			for e in ([self.element] if self.per is None else self.elements.values()):
				e.children[0].disabled = False
				if e.children[0].value: e.children[1].disabled = False
		else: super().disabled(disable)

class CheckgridParam(Param):
	"""A parameter type that takes a list of boolean values. https://helipad.dev/functions/param/"""
	defaultVal = []
	type='checkgrid'

	def __init__(self, **kwargs):
		if kwargs['per'] is not None: raise ValueError(ï('Cannot instantiate per-item checkgrid parameter.'))
		if 'default' not in kwargs or not isinstance(kwargs['default'], list): kwargs['default'] = []
		if isinstance(kwargs['opts'], list): kwargs['opts'] = {k:k for k in kwargs['opts']}
		for k,v in kwargs['opts'].items():
			if isinstance(v, str): kwargs['opts'][k] = (v, None) #Standardized key:(name, tooltip) format
		self.vars = {k: k in kwargs['default'] for k in kwargs['opts']}
		super().__init__(**kwargs)

	@property
	def pKeys(self) -> list: return list(self.vars.keys())

	def getSpecific(self, item=None):
		useElement = self.element is not None and not isNotebook() and not isinstance(self, Shocks)
		vals = self.element if useElement else self.vars
		if item is not None: return vals[item]
		else: return [k for k,v in vals.items() if (v.get() if useElement else v)]

	def set(self, item, val=True, updateGUI: bool=True):
		if getattr(self, 'setter', False):
			val = self.setter(val, item)
			if val is not None: self.setSpecific(item, val)
		else: self.setSpecific(item, val, updateGUI)

	#Takes a list of strings, or a key-bool pair
	def setSpecific(self, item, val=True, updateGUI: bool=True):
		if isinstance(item, list):
			for i in self.pKeys: self.set(i, i in item)
		else:
			if self.element is not None and not isNotebook(): self.element.checks[item].set(val)
			self.vars[item] = val

			if updateGUI and isNotebook() and hasattr(self, 'elements'):
				try: self.elements[item].children[0].value = val
				except KeyError: return #Ipywidgets ≥8.0 runs the callback on instantiation before the element property is set

	@property
	def range(self) -> list:
		combos = []
		for i in range(len(self.vars)): combos += list(combinations(self.pKeys, i+1))
		return combos

	def disabled(self, disable: bool):
		if hasattr(self, 'elements') and isNotebook():
			for e in self.elements.values():
				if hasattr(e, 'children'): e.children[0].disabled = disable
		else: super().disabled(disable)

	def addItem(self, name: str, label: str, position=None, selected: bool=False):
		"""Add a new checkbox to the grid."""
		if self.element is not None and not isNotebook(): raise RuntimeError(ï('Cannot add checkgrid items after control panel is drawn.'))

		if position is None or position > len(self.vars): self.opts[name] = label
		else:		#Reconstruct opts because there's no insert method
			newopts, i = ({}, 1)
			for k,v in self.opts.items():
				if position==i: newopts[name] = label
				newopts[k] = v
				i+=1
			self.opts = newopts

		#Refresh
		if getattr(self, 'containerElement', False):
			from ipywidgets import interactive
			self.elements[name] = interactive(self.setVar(name), val=selected)
			self.elements[name].children[0].description = label
			if position is None or position > len(self.vars): self.containerElement.children += (self.elements[name],)
			else: self.containerElement.children = self.containerElement.children[:position-1] + (self.elements[name],) + self.containerElement.children[position-1:]

		self.vars[name] = selected
		if selected: self.default.append(name)

class ParamGroup:
	"""Define a collapsible group of parameters for display in the control panel. https://helipad.dev/functions/params/group/"""
	def __init__(self, name: str, members, opened: bool):
		self.title = name
		self.members = members
		self.element = None
		self.open = opened

		for p in members.values():
			if p.per is not None or p.type=='checkgrid': raise TypeError(ï('Cannot add checkgrids or per-item parameters to groups.'))
			p.group = name

	def get(self) -> dict:
		"""A `dict` with values for all parameters in the group."""
		return {name: p.get() for name, p in self.members.items()}

	def toggle(self):
		"""Open or close the frame around the grouped parameters."""
		self.open = not self.open

	@property
	def open(self) -> bool:
		"""Whether the frame around the grouped parameters is currently open."""
		return self._open

	@open.setter
	def open(self, val: bool):
		self._open = val
		if self.element is not None:
			if isNotebook(): self.element.selected_index = 0 if val else None
			else: self.element.open = val

#==================
# CONTAINER CLASSES
#==================

class fStoreWithInterface(funcStore):
	"""Subclass of `FuncStore` that also deletes parameter interface elements on removal. https://helipad.dev/functions/funcstore/"""
	def remove(self, name: str):
		if isinstance(name, (list, tuple)): return [self.remove(n) for n in name]

		if name not in self: return False
		if getattr(self[name], 'config', False): raise ValueError(ï('Cannot hide or remove built-in parameters.'))
		self._destroy(self[name])

		del self[name]
		return True

	def hide(self, name: str):
		if isinstance(name, (list, tuple)): return [self.remove(n) for n in name]
		if name not in self or not self[name].element: return False
		if getattr(self[name], 'config', False): raise ValueError(ï('Cannot hide or remove built-in parameters.'))

		if isNotebook():
			if getattr(self[name], 'per', True) or self[name].type=='checkgrid': self[name].element.add_class('helipad_accordion_hidden')
			else: self[name].element.layout.display = 'none'
		else:
			if getattr(self[name], 'per', True) or self[name].type=='checkgrid': self[name].element.pack_forget()
			else: self[name].element.master.pack_forget()

	def show(self, name: str):
		if not self.model.cpanel:
			warnings.warn(ï('Cannot show parameter before control panel is launched.'), None, 2)
			return False

		if isNotebook():
			if getattr(self[name], 'per', True) or self[name].type=='checkgrid': self[name].element.remove_class('helipad_accordion_hidden')
			else: self[name].element.layout.display = None
		else:
			if getattr(self[name], 'per', True) or self[name].type=='checkgrid': self[name].element.pack(fill='x')
			else: self[name].element.master.pack(fill='x')

	def _destroy(self, item):
		if item.element:
			if isNotebook(): item.element.close()
			else:
				#For global non-checkgrid parameters, delete the parent frame. Otherwise item.element is the whole thing
				if getattr(item, 'per', True) or item.type=='checkgrid': item.element.forget()
				else: item.element.master.forget()

class Params(fStoreWithInterface):
	"""Interface for storing, adding, and accessing model parameters. Stored in `model.params`. https://helipad.dev/functions/params/"""
	multi = False

	def __init__(self, model):
		super().__init__()
		self.model = model
		self.groups = []

	def add(self, name: str, title: str, type: str, dflt, opts={}, runtime: bool=True, callback=None, per=None, desc=None, prim=None, getter=None, setter=None, **args):
		"""Register a global parameter to be displayed in the control panel. `type` can take `'menu'`, `'check'`, `'slider'`, `'checkentry'`, `'checkgrid'`, or `'hidden'`, and `opts` will depend on the parameter type. https://helipad.dev/functions/params/add/"""
		if name in self: warnings.warn(ï('Parameter \'{}\' already defined. Overriding…').format(name), None, 2)

		if callable(getter):
			args['getter'] = lambda item=None: getter(*([name, self.model] if per is None else [name, self.model, item]))

		if callable(setter):
			args['setter'] = lambda val, item=None: setter(*([val, name, self.model] if per is None else [val, name, self.model, item]))

		args.update({
			'name': name,
			'title': title,
			'default': dflt,
			'opts': opts,
			'runtime': runtime,
			'callback': callback,
			'desc': desc,
			'per': per
		})
		if per is not None:
			if per=='breed':
				if prim is None:
					if len(list(self.model.agents.keys()))==1: prim = next(iter(self.model.agents.keys()))
					else: raise KeyError(ï('Per-breed parameter must specify which primitive it belongs to.'))
				args['prim'] = prim
			args['pKeys'] = self.model.agents[prim].breeds if per=='breed' else self.model.goods

		if type.title()+'Param' in globals(): pclass = globals()[type.title()+'Param']
		else:
			pclass = Param
			args['type'] = type
		self[name] = pclass(**args)
		if self.model.cpanel and isNotebook(): self.model.cpanel.__init__(self, redraw=True) #Redraw if necessary
		return self[name]

	def group(self, name: str, params: list, opened: bool=True):
		"""Group parameters into a collapsible section in the control panel. https://helipad.dev/functions/params/group/"""
		pg = ParamGroup(name, {p: self[p] for p in params}, opened)
		self.groups.append(pg)
		return pg

	#Save the configuration parameters
	def clear(self):
		"""Clear all user-added parameters and removes them from the control panel if drawn. https://helipad.dev/functions/params/clear/"""
		for p in self.values():
			if getattr(p, 'config', False): continue
			self._destroy(p)
		configs = {k:v for k,v in self.items() if getattr(v, 'config', False)}
		super().clear()
		self.update(configs)

	@property
	def globals(self):
		"""The subset of global parameters, i.e. those added with `per=None`. https://helipad.dev/functions/params/#globals"""
		return {k:v for k,v in self.items() if v.per is None}

	@property
	def perBreed(self):
		"""The subset of per-breed parameters, i.e. those added with `per='breed'`. https://helipad.dev/functions/params/#perbreed"""
		return {k:v for k,v in self.items() if v.per=='breed'}

	@property
	def perGood(self):
		"""The subset of per-good parameters, i.e. those added with `per='good'`. https://helipad.dev/functions/params/#pergood"""
		return {k:v for k,v in self.items() if v.per=='good'}

class Shocks(CheckgridParam, fStoreWithInterface):
	"""Interface for adding, storing, and executing model shocks. Stored in `model.shocks`, but also subclasses `CheckgridParam` and the same object is also located in `model.params['shocks']`. https://helipad.dev/functions/shocks/"""
	multi: bool = False

	def __init__(self, model):
		dict.__init__(self)
		CheckgridParam.__init__(self, name='shocks', title=ï('Shocks'), type='checkgrid', opts={}, dflt={}, runtime=True, config=True, per=None)
		self.model = model

		class Shock(Item):
			"""Container defining a shock to a parameter that can execute during model runtime. https://helipad.dev/functions/shock/"""
			@property
			def selected(self2) -> bool: return self.get(self2.name)

			@selected.setter
			def selected(self2, val: bool): self2.active(val)

			def active(self2, val: bool, updateGUI: bool=True):
				self.set(self2.name, bool(val))
				if updateGUI and not isNotebook() and self2.element is not None:
					self2.element.BooleanVar.set(val)

			def do(self, model):
				if self.param is None: self.valFunc(model)
				else:
					newval = self.valFunc(self.param.get(self.item))	#Pass in current value
					self.param.set(newval, self.item)
					if hasattr(self.param, 'callback') and callable(self.param.callback):
						if self.param.per is not None and self.item is not None:
							self.param.callback(model, self.param.name, self.item, newval)
						else:
							self.param.callback(model, self.param.name, newval)

			#Updates the parameter value when the GUI element is clicked
			#Can't do the regular setVar without a unified Tkinter element (like CheckGrid) to route the values
			def setCallback(self, val=None):
				if not isNotebook(): val = self.element.BooleanVar.get()
				self.active(val, False)

				if callable(model.params['shocks'].callback): model.params['shocks'].callback(model, 'shocks', (self.name, val))
		self.Shock = Shock

	def __repr__(self): return f'<{self.__class__.__name__}: {len(self)} shocks>'

	def add(self, name: str, param, valFunc, timerFunc, active: bool=True, desc=None):
		"""Register a shock to parameter `param`. `valFunc` takes the current value and returns a new value. `timerFunc` is a function that takes the current model time and returns `bool` (or the string `'button'`, in which case the value is shocked when a control panel button is pressed); `valFunc` will execute whenever `timerFunc` returns `True`. `param` can also be set to `None`, in which case `valFunc` receives the model object. https://helipad.dev/functions/shocks/add/"""
		if param is None: item=None
		else:
			if isinstance(param, tuple):
				#Deprecated in Helipad 1.6.1, remove in 1.8
				if len(param) > 1 and param[1] in ['breed', 'good']:
					warnings.warn(ï('Three-item parameter tuple identifiers have been deprecated. The second item can be removed.'), FutureWarning, 2)
					item = param[2]
				else: item = param[1] if len(param)>1 else None
			param = self.model.params[param[0]] if isinstance(param, tuple) else self.model.params[param]

		super().add(name, self.Shock(
			name=name,
			desc=desc,
			param=param,
			item=item,
			valFunc=valFunc,
			timerFunc=timerFunc,
			element=None,
			selected=active
		))

		if timerFunc != 'button': self.addItem(name, name, selected=active)

	def clear(self):
		"""Clear all registered shocks and removes the element from the control panel."""
		if self.element is not None: self._destroy(self)
		super().clear()

	def step(self):
		"""Run through the `timerFunc`s of the registered shocks and execute the `valFunc` for any `timerFunc` that returns `True`. https://helipad.dev/functions/shocks/step/"""
		for shock in self.values():
			if shock.selected and callable(shock.timerFunc) and shock.timerFunc(self.model.t):
				shock.do(self.model)

	@property
	def buttons(self):
		"""The subset of registered shocks where `shock.timerFunc=='button'` and therefore render as buttons in the control panel. https://helipad.dev/functions/shocks/#buttons"""
		return {k:s for k,s in self.items() if s.timerFunc=='button'}

	@property
	def shocksExceptButtons(self):
		"""The subset of registered shocks where `shock.timerFunc` is a function and therefore render as checkboxes in the control panel. https://helipad.dev/functions/shocks/#shocksExceptButtons"""
		return {k:s for k,s in self.items() if callable(s.timerFunc)}

	# ===============
	# TIMER FUNCTIONS
	# The following *return* timer functions; they are not themselves timer functions.
	# ===============

	def randn(self, n):
		"""Generate a timer function that returns `True` with `n`% probability each period. https://helipad.dev/functions/shocks/randn/"""
		if n<0 or n>100: raise ValueError(ï('randn() argument must be between 0 and 100.'))
		def fn(t): return random.randint(0,100) < n
		return fn

	def atperiod(self, n):
		"""Generate a timer function that returns `True` at one or several specified periods. https://helipad.dev/functions/shocks/atperiod/"""
		def fn(t):
			if isinstance(n, list): return t in n
			else: return t==n
		return fn

	def everyn(self, n: int, offset: int=0):
		"""Generates a timer function that returns `True` every `n` periods. https://helipad.dev/functions/shocks/everyn/"""
		def fn(t): return t%n-offset==0
		return fn


# A model of the relative price effects of monetary shocks via helicopter drop
# Download the paper at https://ssrn.com/abstract=2545488

#To generate IRFs, turn off the visualization, turn on stopafter, and one shock.

from itertools import combinations
from math import sqrt
from numpy import random
from helipad import *

M0 = 120000

#===============
# STORE CLASS
# Has to come before adding the primitive
#===============

class Store(baseAgent):
	def __init__(self, breed, aId, model):
		super().__init__(breed, aId, model)

		#Start with equilibrium prices. Not strictly necessary, but it eliminates the burn-in period. See eq. A7
		sm=sum(1/sqrt(model.param(('prod',g))) for g in model.goods.nonmonetary) * M0/(model.param('num_agent')*(len(model.goods.nonmonetary)+sum(1+model.param(('rbd',b)) for b in model.agents['agent'].breeds)))
		self.price = {g:sm/(sqrt(model.param(('prod',g)))) for g in model.goods.nonmonetary}

		self.invTarget = {g:model.param(('prod',g))*model.param('num_agent')*2 for g in model.goods.nonmonetary}
		self.portion = {g:1/(len(model.goods.nonmonetary)) for g in model.goods.nonmonetary} #Capital allocation
		self.wage = 0
		self.cashDemand = 0

	def step(self, stage):
		super().step(stage)
		N = self.model.param('num_agent')

		#Calculate wages
		self.cashDemand = N * self.wage #Hold enough cash for one period's disbursements
		newwage = (self.balance - self.cashDemand) / N
		newwage = max(newwage, 1)
		self.wage = (self.wage * self.model.param('wStick') + newwage)/(1 + self.model.param('wStick'))
		if self.wage * N > self.balance: self.wage = self.balance / N 	#Budget constraint

		#Hire labor, with individualized wage shocks
		labor = 0
		for a in self.model.agents['agent']:
			self.wage = max(self.wage, 0)
			wage = random.normal(self.wage, self.wage/2 + 0.1)	#Can't have zero stdev
			wage = max(wage, 0)									#Wage bounded from below by 0
			self.pay(a, wage)
			labor += 1

		tPrice = sum(self.price[good] for good in self.model.goods.nonmonetary)
		avg, stdev = {},{} #Hang onto these for use with credit calculations
		for i in self.model.goods.nonmonetary:

			#Just have a fixed inventory target, but update if params do
			self.invTarget = {g:self.model.param(('prod',g))*self.model.param('num_agent')*2 for g in self.model.goods.nonmonetary}

			#Produce stuff
			self.portion[i] = (self.model.param('kImmob') * self.portion[i] + self.price[i]/tPrice) / (self.model.param('kImmob') + 1)	#Calculate capital allocation
			self.stocks[i] = self.stocks[i] + self.portion[i] * labor * self.model.param(('prod', i))

			#Set prices
			#Change in the direction of hitting the inventory target
			# self.price[i] += log(self.invTarget[i] / (self.inventory[i][0] + self.lastShortage[i])) #Jim's pricing rule?
			self.price[i] += (self.invTarget[i] - self.stocks[i] + self.model.data.getLast('shortage-'+i))/100 #/150

			#Adjust in proportion to the rate of inventory change
			#Positive deltaInv indicates falling inventory; negative deltaInv rising inventory
			lasti = self.model.data.getLast('inv-'+i,2)[0] if self.model.t > 1 else 0
			deltaInv = lasti - self.stocks[i]
			self.price[i] *= (1 + deltaInv/(50 ** self.model.param('pSmooth')))
			if self.price[i] < 0: self.price[i] = 1

#===============
# CONFIGURATION
#===============

#Stick this here so we can build on it with the OMO model
def setup():
	heli = Helipad()
	heli.agents.addPrimitive('store', Store, dflt=1, priority=2, hidden=True)

	# Configure how many breeds there are and what good each consumes
	# In this model, goods and breeds correspond, but they don't necessarily have to
	breeds = [
		('hobbit', 'jam', '#D73229'),
		('dwarf', 'axe', '#2D8DBE'),
		# ('elf', 'lembas', '#CCBB22')
	]
	AgentGoods = {}
	for b in breeds:
		heli.agents.addBreed(b[0], b[2], prim='agent')
		heli.goods.add(b[1], b[2])
		AgentGoods[b[0]] = b[1] #Hang on to this list for future looping
	heli.goods.add('cash', '#009900', money=True)

	heli.name = 'Helicopter'
	heli.agents.order = 'random'

	# UPDATE CALLBACKS

	def ngdpUpdater(model, var, val):
		if model.hasModel: model.cb.ngdpTarget = val if not val else model.cb.ngdp

	def rbalUpdater(model, var, breed, val):
		if model.hasModel:
			beta = val/(1+val)

			for a in model.agents['agent']:
				if hasattr(a, 'utility') and a.breed == breed:
					a.utility.coeffs['rbal'] = beta
					a.utility.coeffs['good'] = 1-beta

	heli.params.add('ngdpTarget', 'NGDP Target', 'check', dflt=False, callback=ngdpUpdater)
	heli.params.add('dist', 'Distribution', 'menu', dflt='prop', opts={
		'prop': 'Proportional',
		'lump': 'Lump Sum'
	}, runtime=False)

	heli.params.add('pSmooth', 'Price Smoothness', 'slider', dflt=1.5, opts={'low': 1, 'high': 3, 'step': 0.05})
	heli.params.add('wStick', 'Wage Stickiness', 'slider', dflt=10, opts={'low': 1, 'high': 50, 'step': 1})
	heli.params.add('kImmob', 'Capital Immobility', 'slider', dflt=100, opts={'low': 1, 'high': 150, 'step': 1})
	heli.params.group('Environment', ('pSmooth', 'wStick', 'kImmob'), opened=False)

	#Low Es means the two are complements (0=perfect complements)
	#High Es means the two are substitutes (infinity=perfect substitutes)
	#Doesn't really affect anything though – even utility – so don't bother exposing it
	heli.params.add('sigma', 'Elast. of substitution', 'hidden', dflt=.5, opts={'low': 0, 'high': 10, 'step': 0.1})

	heli.params.add('rbd', 'Demand for Real Balances', 'slider', per='breed', dflt={'hobbit':7, 'dwarf': 35}, opts={'low':1, 'high': 50, 'step': 1}, prim='agent', callback=rbalUpdater)
	heli.params.add('prod', 'Productivity', 'slider', per='good', dflt=1.75, opts={'low':0.1, 'high': 2, 'step': 0.1}) #If you shock productivity, make sure to call rbalupdater

	#Takes as input the slider value, outputs b_g. See equation (A8) in the paper.
	def rbaltodemand(breed):
		def reporter(model):
			rbd = model.param(('rbd', breed))
			beta = rbd/(1+rbd)

			return (beta/(1-beta)) * len(model.goods) * sqrt(model.param(('prod',AgentGoods[breed]))) / sum(1/sqrt(pr) for pr in model.param('prod').values())

		return reporter

#===============
# DATA COLLECTION AND VISUALIZATION
#===============

	from helipad.visualize import TimeSeries
	viz = heli.useVisual(TimeSeries)
	viz.dim = (1000, 800)
	viz.pos = (400, 100)

	viz.addPlot('prices', 'Prices', 1, selected=True)
	viz.addPlot('inventory', 'Inventory', 3)
	viz.addPlot('rbal', 'Real Balances', 5)
	viz.addPlot('ngdp', 'NGDP', 7, selected=False)
	viz.addPlot('capital', 'Production', 9, selected=False)
	viz.addPlot('wage', 'Wage', 11, selected=False)
	viz.addPlot('shortage', 'Shortages', 6, selected=False)

	viz['capital'].addSeries(lambda t: 1/len(heli.agents['agent'].breeds), '', '#CCCCCC')

	# heli.data.addReporter('rBal', heli.data.agentReporter('realBalances', 'agent', breed=True))
	def shortageReporter(good):
		def reporter(model): return model.shortages[good]
		return reporter

	for breed, d in heli.agents['agent'].breeds.items():
		heli.data.addReporter('rbalDemand-'+breed, rbaltodemand(breed))
		heli.data.addReporter('shortage-'+AgentGoods[breed], shortageReporter(AgentGoods[breed]))
		heli.data.addReporter('eCons-'+breed, heli.data.agentReporter('expCons', 'agent', breed=breed, stat='sum'))
		# heli.data.addReporter('rWage-'+breed, lambda model: heli.data.agentReporter('wage', 'store')(model) / heli.data.agentReporter('price', 'store', good=b.good)(model))
		# heli.data.addReporter('expWage', heli.data.agentReporter('expWage', 'agent'))
		heli.data.addReporter('rBal-'+breed, heli.data.agentReporter('realBalances', 'agent', breed=breed))
		heli.data.addReporter('invTarget-'+AgentGoods[breed], heli.data.agentReporter('invTarget', 'store', good=AgentGoods[breed]))
		heli.data.addReporter('portion-'+AgentGoods[breed], heli.data.agentReporter('portion', 'store', good=AgentGoods[breed]))

		viz['demand'].addSeries('eCons-'+breed, breed.title()+'s\' Expected Consumption', d.color2)
		viz['shortage'].addSeries('shortage-'+AgentGoods[breed], AgentGoods[breed].title()+' Shortage', d.color)
		viz['rbal'].addSeries('rbalDemand-'+breed, breed.title()+' Target Balances', d.color2)
		viz['rbal'].addSeries('rBal-'+breed, breed.title()+ 'Real Balances', d.color)
		viz['inventory'].addSeries('invTarget-'+AgentGoods[breed], AgentGoods[breed].title()+' Inventory Target', heli.goods[AgentGoods[breed]].color2)
		viz['capital'].addSeries('portion-'+AgentGoods[breed], AgentGoods[breed].title()+' Capital', heli.goods[AgentGoods[breed]].color)
		# viz['wage'].addSeries('expWage', 'Expected Wage', '#999999')

	#Do this one separately so it draws on top
	for good, g in heli.goods.nonmonetary.items():
		heli.data.addReporter('inv-'+good, heli.data.agentReporter('stocks', 'store', good=good))
		viz['inventory'].addSeries('inv-'+good, good.title()+' Inventory', g.color)
		heli.data.addReporter('price-'+good, heli.data.agentReporter('price', 'store', good=good))
		viz['prices'].addSeries('price-'+good, good.title()+' Price', g.color)

	#Price ratio plots
	def ratioReporter(item1, item2):
		def reporter(model):
			return model.data.agentReporter('price', 'store', good=item1)(model)/model.data.agentReporter('price', 'store', good=item2)(model)
		return reporter
	viz.addPlot('ratios', 'Price Ratios', position=3, logscale=True)
	viz['ratios'].addSeries(lambda t: 1, '', '#CCCCCC')	#plots ratio of 1 for reference without recording a column of ones

	for r in combinations(heli.goods.nonmonetary.keys(), 2):
		heli.data.addReporter('ratio-'+r[0]+'-'+r[1], ratioReporter(r[0], r[1]))
		c1, c2 = heli.goods[r[0]].color, heli.goods[r[1]].color
		viz['ratios'].addSeries('ratio-'+r[0]+'-'+r[1], r[0].title()+'/'+r[1].title()+' Ratio', c1.blend(c2))

	#Misc plots
	heli.data.addReporter('ngdp', lambda model: model.cb.ngdp)
	viz['ngdp'].addSeries('ngdp', 'NGDP', '#000000')
	heli.data.addReporter('storeCash', heli.data.agentReporter('balance', 'store'))
	viz['money'].addSeries('storeCash', 'Store Cash', '#777777')
	heli.data.addReporter('StoreCashDemand', heli.data.agentReporter('cashDemand', 'store'))
	viz['money'].addSeries('StoreCashDemand', 'Store Cash Demand', '#CCCCCC')
	heli.data.addReporter('wage', heli.data.agentReporter('wage', 'store'))
	viz['wage'].addSeries('wage', 'Wage', '#000000')

#================
# POST-ANALYSIS
# Generate the appropriate impulse response functions
#================

	@heli.hook
	def terminate(model, data):
		shocks = model.params['shocks'].get()
		if not model.param('stopafter') or not shocks: return

		from statsmodels.tsa.api import VAR
		import matplotlib.pyplot as plt

		if 'Dwarf' in shocks[0]:	#Equilibrating shocks
			data = data[['M0', 'ratio-jam-axe', 'rBal-dwarf']]
		elif 'M0' in shocks[0]:		#Disequilibrating shocks
			data = data[['M0', 'ratio-jam-axe']]

		model = VAR(data)
		results = model.fit(50)
		irf = results.irf(50)
		irf.plot(orth=False, impulse='M0', response='ratio-jam-axe')
		plt.show(block=False)

#================
# AGENT BEHAVIOR
#================

	from helipad.utility import CES

	@heli.hook
	def agentInit(agent, model):
		agent.store = model.agents['store'][0]
		agent.item = AgentGoods[agent.breed]
		rbd = model.param(('rbd', agent.breed))
		beta = rbd/(rbd+1)
		agent.utility = CES({'good': 1-beta, 'rbal': beta }, agent.model.param('sigma'))
		agent.expCons = model.param(('prod', agent.item))

		#Set cash endowment to equilibrium value based on parameters. Not strictly necessary but avoids the burn-in period.
		agent.stocks[model.goods.money] = agent.store.price[agent.item] * rbaltodemand(agent.breed)(heli)

	@heli.hook
	def agentStep(agent, model, stage):
		itemPrice = agent.store.price[agent.item]

		q = agent.utility.demand(agent.balance, {'good': itemPrice, 'rbal': itemPrice})['good']	#Equimarginal condition given CES between real balances and consumption
		basicq = q						#Save this for later since we adjust q
		bought = agent.buy(agent.store, agent.item, q, itemPrice)
		if bought < q: model.shortages[agent.item] += q-bought #Record shortages
		if agent.stocks[model.goods.money] < 0: agent.stocks[model.goods.money] = 0	#Floating point error gives infinitessimaly negative cash sometimes
		agent.utils = agent.utility.calculate({'good': agent.stocks[agent.item], 'rbal': agent.balance/itemPrice}) if hasattr(agent,'utility') else 0	#Get utility
		agent.stocks[agent.item] = 0 #Consume goods

		negadjust = q - bought											#Update your consumption expectations if the store has a shortage
		negadjust = min(negadjust, basicq)
		agent.expCons = (19 * agent.expCons + basicq-negadjust)/20		#Set expected consumption as a decaying average of consumption history

	def realBalances(agent):
		if not hasattr(agent, 'store'): return 0
		return agent.balance/agent.store.price[agent.item]
		# return agent.balance/agent.model.cb.P
	Agent.realBalances = property(realBalances)

	@heli.hook
	def modelPreStep(model):
		model.shortages = {g:0 for g in model.goods.nonmonetary}

	@heli.hook
	def modelPostSetup(model): model.cb = CentralBank(0, model)

	@heli.hook
	def modelPostStep(model): model.cb.step()	#Step the central bank last

#========
# SHOCKS
#========

	#Random shock to dwarf cash demand
	def shock(v):
		c = random.normal(v, 4)
		return c if c >= 1 else 1
	heli.shocks.add('Dwarf real balances', ('rbd','dwarf'), shock, heli.shocks.randn(2), active=False)

	#Shock the money supply
	def mshock(model):
		pct = random.normal(3, 15) #High mean to counteract the downward bias of (1-%)(1+%)
		m = model.cb.M0 * (1+pct/100)
		m = max(m, 10000)		#Things get weird when there's a money shortage
		model.cb.M0 = m
	heli.shocks.add('M0 (2% prob)', None, mshock, heli.shocks.randn(2), desc="Shocks the money supply a random percentage (µ=1, σ=15) with 2% probability each period")

	#Only one shock at a time
	def shocksCallback(model, var, val):
		shocks = heli.param('shocks')
		if len(shocks) > 1:
			for s in shocks:
				if s != val[0]:
					model.shocks[s].active(False)

		if 'M0' in val[0] and val[1]: model.param('ngdpTarget', False)
		elif 'Dwarf' in val[0] and val[1]: model.param('ngdpTarget', True)
	heli.params['shocks'].callback = shocksCallback

	return heli

#
# Central Bank
#

class CentralBank(baseAgent):
	ngdpAvg = 0
	ngdp = 0
	primitive = 'cb'

	def __init__(self, aId, model):
		super().__init__(None, aId, model)
		self.id = aId
		self.model = model
		self.stocks[self.model.goods.money] = M0 #Has to have assets in order to contract

		self.ngdpTarget = 10000 if model.param('ngdpTarget') else False

	def step(self):

		#Record macroeconomic vars at the end of the last stage
		#Getting demand has it lagged one period…
		self.ngdp = sum(self.model.data.getLast('demand-'+good) * self.model.agents['store'][0].price[good] for good in self.model.goods.nonmonetary)
		if not self.ngdpAvg: self.ngdpAvg = self.ngdp
		self.ngdpAvg = (2 * self.ngdpAvg + self.ngdp) / 3

		#Set macroeconomic targets
		expand = 0
		if self.ngdpTarget: expand = self.ngdpTarget - self.ngdpAvg
		if expand != 0: self.expand(expand)

	def expand(self, amount):
		if self.model.param('dist') == 'lump':
			amt = amount/self.model.param('num_agent')
			for a in self.model.agents['agent']:
				a.stocks[self.model.goods.money] += amt
		else:
			for a in self.model.agents.all:
				a.stocks[self.model.goods.money] += a.stocks[self.model.goods.money]/self.M0 * amount

	@property
	def M0(self):
		return self.model.data.agentReporter('stocks', 'all', good=self.model.goods.money, stat='sum')(self.model)

	@M0.setter
	def M0(self, value): self.expand(value - self.M0)

#Only launch the cpanel if we haven't embedded
if __name__ == '__main__':
	heli = setup()
	heli.launchCpanel()


#A reconstruction of Hurford (1991), "The Evolution of the Critical Period for Language Acquisition"
#https://www.sciencedirect.com/science/article/abs/pii/001002779190024X

from helipad import Helipad, Agent
import random
from numpy.random import choice
from numpy import mean

heli = Helipad()
heli.name = 'Critical Period'
heli.agents.order = 'random'

heli.params.add('pleio', 'Pleiotropy', 'slider', dflt=7, opts={'low': 0, 'high': 10, 'step': 1}, runtime=False)
heli.params.add('acq', 'Acquisition from', 'menu', dflt='mother', opts={'mother': 'Mother', 'wholepop': 'Whole Population'})
heli.params.add('condition', 'Language affects', 'menu', dflt='rep', opts={'rep': 'Reproduction', 'srv': 'Survival'})

heli.params['num_agent'].opts['step'] = 10 #Need it to be divisible by the number of life stages
heli.params['num_agent'].opts['low'] = 10
heli.param('num_agent', 30)

#Returns language capacity
def capacity(self, gene=None, age=None):
	if gene is None: gene = self.dominantLap
	if age is None: age=self.age
	
	start = (10-self.model.param('pleio'))*age
	l = sum(gene[start:start+10])
	return max(0,l)
Agent.capacity = capacity

@heli.hook
def modelPostSetup(model):
	model.births = 0 #Keep track so we can mutate every 30
	
#Initialize agent variables
@heli.hook
def agentInit(agent, model):
	agent.language = 0
	agent.dominantLap = [2*i%2-1 for i in range(100-9*model.param('pleio'))] #±1
	agent.recessivLap = [2*i%2-1 for i in range(100-9*model.param('pleio'))]
	agent.hasReproduced = False
	agent.age = agent.id%10

@heli.reporter
def adultLanguage(model):
	return mean([a.language for a in model.agents['agent'] if a.age>1])

@heli.hook
def modelPreStep(model):
	for a in model.agents['agent']: a.hasReproduced = False

#Select people to die in accidents
@heli.hook
def modelPostStep(model):
	if model.param('condition')=='srv':
		p = [((a.age+1)/(a.language+1))**3 for a in model.agents['agent']]
		p = [n/sum(p) for n in p] #Normalize probabilities to add up to 1
		for a in choice(model.agents['agent'], int(model.param('num_agent')/10), False, p): a.die() #Paper says "up to 3" die per period, but doesn't say how the "up to" is determined

#Learn language and die if old
@heli.hook
def agentStep(agent, model, stage):
	capacity = agent.capacity()
	if model.param('acq') == 'mother' or not agent.age:
		mother = agent.parent[0] if isinstance(agent.parent, list) else agent.parent #Dead agents are automatically removed from parent lists
		exemp = mother.language if mother is not None else 0
	else: exemp = adultLanguage(model)
	
	agent.language += capacity if exemp >= agent.language + capacity else (capacity + exemp-agent.language)/2
			
	if agent.age >= 9: agent.die()

#Replace dead agents
@heli.hook
def agentDie(agent):
	
	#Nominate parents with probability proportional to the cube of the language score, in the reproduction condition
	#Or with equal probability in the survival condition
	if heli.param('condition')=='rep':
		nominees = []
		for a in heli.agents['agent']:
			if a.hasReproduced or a.age<=1: continue
			for n in range(round(a.language**3+1)): nominees.append(a.id) #+1 because otherwise we start out with no one able to reproduce
	else: nominees = [a.id for a in heli.agents['agent'] if not a.hasReproduced and a.age>1]
	n1 = n2 = heli.agents[random.choice(nominees)]
	while n1.id == n2.id: n2 = heli.agents[random.choice(nominees)]
	
	#Complicated inheritance so do it manually
	baby = n1.reproduce(partners=[n2])
	parents = [n1, n2]
	for n in parents:
		n.gamete = []
		inp = [n.dominantLap, n.recessivLap]
		for i,v in enumerate(inp[1]):
			n.gamete.append(inp[random.randrange(len(inp))][i])
	mf = random.randrange(len(parents))
	baby.dominantLap = parents[mf].gamete
	baby.recessivLap = parents[not mf].gamete
	heli.births += 1
	
	#Mutate by flipping a bit in the dominant LAP once every 30 births
	if not heli.births%30:
		r = random.randrange(len(baby.dominantLap))
		baby.dominantLap[r] = -baby.dominantLap[r]

# @heli.event(repeat=True)
# def test(model): return model.t%100==95

#Visualization
from helipad.visualize import TimeSeries, Charts
viz = heli.useVisual(Charts)
# lplot = viz.addPlot('language', 'Language')
# lplot.addSeries('adultLanguage', 'Adult Language', 'blue')

# gplot = viz.addPlot('geno', 'Genotypes')
gchart = viz.addPlot('geno', 'Capacity by stage (dominant allele)')
gchart2 = viz.addPlot('rec', 'Capacity by stage (recessive allele)', horizontal=True)

def genoReporter(age, gene=False):
	def rep(model):
		return mean([a.capacity(gene=None if not gene else a.recessivLap, age=age) for a in model.agents['agent']])
	return rep
gcolors = ['F00', 'F03', 'F06', 'F09', 'F0C', 'C0F', '90F', '60F', '30F', '00F']
for age in range(10):
	heli.data.addReporter('geno-'+str(age), genoReporter(age))
	heli.data.addReporter('rec-'+str(age), genoReporter(age, gene='rec'))
	# gplot.addSeries('geno-'+str(age), 'Capacity age '+str(age), '#'+gcolors[age])
	gchart.addBar('geno-'+str(age), str(age), '#'+gcolors[age])
	gchart2.addBar('rec-'+str(age), str(age), '#'+gcolors[age])

heli.launchCpanel()


"""
The Tkinter-based control panel class for use in standalone models. This module should not be imported directly; use `model.launchCpanel()` instead.
"""

import tkinter as tk
import os, sys
from math import ceil
from tkinter.ttk import Progressbar
from helipad.helpers import Color, ï
from helipad.param import Param

class Cpanel(tk.Tk):
	"""The Tkinter-based control panel class for use in standalone models. https://helipad.dev/functions/cpanel/"""
	font = ('Lucida Grande', 16) if sys.platform=='darwin' else ('Calibri', 14)

	def __init__(self, model):
		self.model = model
		super().__init__()
		self.protocol("WM_DELETE_WINDOW", self.cpanelClose)

		#Set application name
		self.setAppIcon()
		self.title(ï('{}Control Panel').format(self.model.name+(' ' if self.model.name!='' else '')))
		self.resizable(0,0)

		bgcolors = ('#FFFFFF','#EEEEEE')
		fnum = 1

		#
		# CALLBACK FUNCTION GENERATORS FOR TKINTER ELEMENTS
		#
		def setVar(param, item=None):
			def sv(val=None):
				#Different widgets send different things to the callback…
				if param.type=='slider': val = float(val)
				elif param.type=='menu': val = {y:x for x,y in param.opts.items()}[val]
				elif param.type=='check': val = (param.element if item is None else param.elements[item]).BooleanVar.get()

				#Parameters that don't update automatically
				if param.type in ['slider', 'menu', 'check']: param.set(val, item, updateGUI=False)

				if callable(param.callback):
					if param.per is None: param.callback(self.model, param.name, val)
					else: param.callback(self.model, param.name, item, val)
			return sv
		Param.setVar = setVar

		#For shock buttons.
		#Can't do an inline lambda here because lambdas apparently don't preserve variable context
		def shockCallback(name: str):
			return lambda: self.model.shocks[name].do(self.model)

		class progressBar(Progressbar):
			def __init__(self, determinate: bool=True, root=None):
				super().__init__(root, length=250, style="whitebg.Horizontal.TProgressbar")
				self.determinate(determinate, False)
				self.running = False

			@property
			def mode(self):
				#Windows returns a string here, and MacOS returns an object
				mode = self.cget('mode')
				return mode if isinstance(mode, str) else self.cget('mode').string

			def determinate(self2, det: bool, refresh: bool=True):
				self2.config(mode='determinate' if det else 'indeterminate')
				if det: super().stop()
				elif self2.running: super().start()
				if refresh: self.update()
			def update(self, n): self['value'] = n*100
			def start(self):
				if self.mode =='indeterminate': super().start()
			def stop(self):
				if self.mode =='indeterminate':
					super().stop()
					self.update(1)
			def done(self):
				self.stop()
				self.config(mode='determinate')
				self.update(0)

		class runButton(tk.Button):
			def __init__(self2, root, bg='#EEEEEE'):
				super().__init__(root, text='Run', command=self.model.launchVisual, padx=10, pady=10, highlightbackground=bg)

			def run(self2):
				self2['text'] = ï('Pause')
				self2['command'] = self.model.stop

			def pause(self2):
				self2['text'] = ï('Run')
				self2['command'] = self.model.start

			def terminate(self2):
				self2['text'] = ï('New Model')
				self2['command'] = self.model.launchVisual

		#
		# CONSTRUCT CONTROL PANEL INTERFACE
		#

		self.tiers = { 'config': tk.Frame(self, padx=10, pady=10, bg=bgcolors[fnum%2]) }

		ctop = self.model.doHooks('CpanelTop', [self, bgcolors[fnum%2]])
		if ctop:
			ctop.pack(fill='x', side='top')
			fnum += 1

		self.renderParam(self.model.params['stopafter']).grid(row=0,column=0, columnspan=3, sticky='w')
		self.renderParam(self.model.params['csv']).grid(row=1,column=0, columnspan=3, sticky='w')
		if not self.model.params['stopafter'].event and not self.model.param('stopafter'): self.model.params['stopafter'].element.entryValue.set(10000)
		self.model.params['csv'].set(ï('filename'))
		self.model.params['csv'].set(False)

		self.renderParam(self.model.params['refresh']).grid(row=2, column=0, columnspan=2, pady=(10,0))
		self.runButton = runButton(self.tiers['config'])
		self.runButton.grid(row=2, column=2, pady=(15,0))

		for c in range(2): self.tiers['config'].columnconfigure(c, weight=1)
		self.tiers['config'].pack(fill='x', side='top')
		fnum += 1

		#Can't change the background color of a progress bar on Mac, so we have to put a gray stripe on top :-/
		frame0 = tk.Frame(self, padx=10, pady=0, bg=bgcolors[1])
		self.progress = progressBar(root=frame0)
		self.progress.grid(row=0, column=0)
		frame0.columnconfigure(0,weight=1)
		frame0.pack(fill='x', side='top')

		caip = self.model.doHooks('CpanelAboveItemParams', [self, bgcolors[fnum%2]])
		if caip:
			caip.pack(fill='x', side='top')
			fnum += 1

		#Tiers for per-good and per-breed parameters
		self.tiers['pergood'] = tk.Frame(self, bg=bgcolors[fnum%2])
		self.tiers['pergood'].pack(fill='x', side='top')
		if model.params.perGood: fnum += 1 #Only increment the stripe counter if we had any good params to draw
		self.tiers['perbreed'] = tk.Frame(self, bg=bgcolors[fnum%2])
		self.tiers['perbreed'].pack(fill='x', side='top')
		if model.params.perBreed: fnum += 1

		cap = self.model.doHooks('CpanelAboveParams', [self, bgcolors[fnum%2]])
		if cap:
			cap.pack(fill='x', side='top')
			fnum += 1

		#Set up tiers for global, grouped, and checkgrid parameters
		self.tiers['global'] = tk.Frame(self, bg=bgcolors[fnum%2])
		self.tiers['global'].pack(fill='x', side='top')
		if model.params.globals: fnum += 1

		for group in self.model.params.groups:
			group.element = expandableFrame(self, bg=bgcolors[fnum%2], padx=5, pady=10, text=group.title, fg="#333", font=self.font, startOpen=group.open)
			if group.element is not None: group.element.pack(fill='x')
			fnum += 1

		self.tiers['checkgrids'] = tk.Frame(self, bg=bgcolors[fnum%2])
		self.tiers['checkgrids'].pack(fill='x', side='top')
		if [p for p in self.model.params.values() if p.type=='checkgrid' and p.name!='shocks']: fnum += 1

		#Fill in tiers with parameters
		for k, param in [(k,p) for k,p in self.model.params.items() if not getattr(p, 'config', False) or p.name=='plots']:
			e = self.renderParam(param)
			if e is not None: e.pack(fill='x')

		cas = self.model.doHooks('CpanelAboveShocks', [self, bgcolors[fnum%2]])
		if cas:
			cas.pack(fill='x', side='top')
			fnum += 1

		#Shock checkboxes and buttons
		if len(self.model.shocks):
			self.model.shocks.element = expandableFrame(self, text=self.model.shocks.title, padx=5, pady=8, font=self.font, bg=bgcolors[fnum%2])
			self.model.shocks.element.checks = {}
			self.model.params['shocks'].element = self.model.shocks.element
			for shock in self.model.shocks.shocksExceptButtons.values():
				bv = tk.BooleanVar(value=shock.selected)
				shock.element = tk.Checkbutton(self.model.shocks.element.subframe, text=shock.name, var=bv, onvalue=True, offvalue=False, bg=bgcolors[fnum%2], anchor='w', command=shock.setCallback)
				shock.element.BooleanVar = bv #To set via the shock object
				self.model.shocks.element.checks[shock.name] = bv #To set via the shock parameter
				if shock.desc is not None: Tooltip(shock.element, shock.desc)
				shock.element.pack(fill='both')

			b=0
			if len(self.model.shocks.buttons):
				bframe = tk.Frame(self.model.shocks.element.subframe, bg=bgcolors[fnum%2])
				for c in range(2): bframe.columnconfigure(c ,weight=1)
				for shock in self.model.shocks.buttons.values():
					shock.element = tk.Button(bframe, text=shock.name, command=shockCallback(shock.name), padx=10, pady=10, highlightbackground=bgcolors[fnum%2])
					shock.element.grid(row=3+int(ceil((b+1)/2)), column=b%2, sticky='w')
					if shock.desc is not None: Tooltip(shock.element, shock.desc)
					b+=1
				bframe.pack(fill='both')
			self.model.shocks.element.pack(fill='x', side='top')

		cbot = self.model.doHooks('CpanelBottom', [self, bgcolors[fnum%2]])
		if cbot:
			cbot.pack(fill='x', side='top')
			fnum += 1

	def renderParam(self, param, item=None, frame=None):
		"""Constructs a Tkinter widget from a `Param` object. https://helipad.dev/functions/cpanel/renderparam/"""
		if param.type=='hidden': return

		#Parent frame for per-item parameters
		if param.per is not None and item is None:
			param.element = expandableFrame(self.tiers['per'+param.per], bg=self.tiers['per'+param.per].cget('bg'), padx=5, pady=10, text=param.title, fg="#333", font=self.font)
			efSub = param.element.subframe
			i=0
			param.elements = {}
			for name, b in param.pKeys.items():
				if getattr(b, 'money', False): continue

				#Do this separately because they should all be stacked
				f = self.renderParam(param, item=name, frame=efSub)
				if param.type == 'checkentry':
					f.grid(row=i, column=0)
					efSub.columnconfigure(0,weight=1)

				#Everything else goes in the two-column format
				else:
					f.grid(row=ceil((i+1)/2)*2, column=i%2)
					for c in range(2): efSub.columnconfigure(c, weight=1)

				i+=1
			return param.element

		elif param.type=='checkgrid':
			param.element = checkGrid(self.tiers['checkgrids'], text=param.title, columns=getattr(param, 'columns', 3), bg=self.tiers['checkgrids'].cget('bg'), callback=param.setVar())
			for k,v in param.opts.items():
				if not isinstance(v, (tuple, list)): v = (v, None)
				elif len(v) < 2: v = (v[0], None)
				param.element.addCheck(k, v[0], param.vars[k], v[1])
			return param.element

		#Single parameters, including the individual per-item parameters
		else:
			def drawCircle(frame, color):
				circle = tk.Canvas(frame, width=17, height=12, bg=frame.cget('bg'), highlightthickness=0)
				circle.create_oval(0,0,12,12,fill=color, outline='')
				return circle

			title = param.title if item is None else item.title()

			#Find the right frame to attach it to
			if getattr(param,'config',False): frame = self.tiers['config']
			elif frame is None:
				frame = self.tiers['global']
				for group in self.model.params.groups:
					if param.name in group.members:
						frame = group.element.subframe
						break
			wrap = tk.Frame(frame, bg=frame.cget('bg'), padx=10 if item is None and not getattr(param,'config',False) else 0, pady=8 if item is None and not getattr(param,'config',False) else 0)

			#Get .value directly rather than .get because we need the Var() items
			#Except for checkentry since it doesn't store its values in .value
			if param.value is not None:
				val = param.value if item is None else param.value[item]

			#These put the circle beside the widget
			if param.type in ['check', 'checkentry']:
				if param.type=='check':
					v = tk.BooleanVar(value=val)
					el = tk.Checkbutton(wrap, text=title, var=v, onvalue=True, offvalue=False, command=param.setVar(item), bg=wrap.cget('bg'))
					el.BooleanVar = v
				elif param.type=='checkentry':
					dflt = param.get(item)
					el = checkEntry(wrap, title, bg=wrap.cget('bg'), width=15, padx=0 if getattr(param,'config',False) else 10, pady=0 if getattr(param,'config',False) else 5, datatype='int' if param.entryType is int else 'string', command=param.setVar(item))
					if param.name=='stopafter' and param.event:
						el.disable()
						el.entryValue.set('Event: '+param.get())
						el.checkVar.set(True)
						el.textbox.config(font=('Helvetica Neue', 12,'italic')) #Lucida doesn't have an italic?
					else: el.set(dflt)

				if item is not None:
					el.grid(row=0, column=1)
					drawCircle(wrap, param.pKeys[item].color.hex).grid(row=0, column=0)
				else: el.pack(anchor='center' if param.type=='check' else 'w')

			#These need a separate label
			else:
				if param.type == 'menu':
					v = tk.StringVar(value=param.opts[val])
					el = tk.OptionMenu(wrap, v, *param.opts.values(), command=param.setVar(item))
					el.StringVar = v #Save to set later
					el.config(bg=wrap.cget('bg'))
				elif param.type == 'slider':
					if isinstance(param.opts, dict): el = tk.Scale(wrap, from_=param.opts['low'], to=param.opts['high'], resolution=param.opts['step'], orient='horizontal', length=150, highlightthickness=0, command=param.setVar(item), bg=wrap.cget('bg'))
					else: el = logSlider(wrap, title=title if getattr(param, 'config', False) else None, orient='horizontal', values=param.opts, length=150, command=param.setVar(item), bg=wrap.cget('bg'))
					el.set(param.get(item))

				if item is None and not getattr(param, 'config', False):
					tk.Label(wrap, text=title, fg="#333", bg=wrap.cget('bg')).pack(side='left', padx=8, pady=3)
					el.pack(side='right')
				elif getattr(param, 'config', False): el.pack()
				else:
					lframe = tk.Frame(wrap, bg=wrap.cget('bg'), padx=0, pady=0)
					tk.Label(lframe, text=title, fg="#333", bg=lframe.cget('bg')).grid(row=0, column=1, pady=(0,8))
					drawCircle(lframe, param.pKeys[item].color.hex).grid(row=0, column=0, pady=(0,8))
					lframe.grid(row=1, column=0)
					el.grid(row=0,column=0)

			if param.desc is not None: Tooltip(wrap, param.desc)

			if item is None: param.element = el
			else: param.elements[item] = el
			return wrap

	#Separate function so we can call it again when MPL tries to override
	def setAppIcon(self):
		"""Set the dock/taskbar icon to the Helipad icon."""
		try:
			__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
			icon = os.path.join(__location__, 'Helipad.png')
			pi = tk.PhotoImage(file=icon, master=self)
			self.tk.call('wm','iconphoto', self._w, pi)
		except: pass

	def cpanelClose(self):
		"""Cleanup on cpanel close so a running model will keep running."""
		for p in self.model.params.values(): p.element = None
		self.model.cpanel = None
		self.destroy()

	def step(self) -> int:
		"""Step one period at a time and update the visualization. For use in debugging."""
		t = self.model.step()
		self.model.graph.update(self.model.data.getLast(1))
		return t

#
# MISCELLANEOUS INTERFACE ELEMENTS
#

class logSlider(tk.Frame):
	"""A slider with discrete specified intervals. https://helipad.dev/functions/logslider/"""
	def __init__(self, parent=None, title=None, command=None, bg='#FFFFFF', font=('Lucida Grande',12), values=(), **kwargs):
		tk.Frame.__init__(self, parent, bg=bg)
		self.label = tk.Label(self, font=font, text=title, bg=bg).pack(side='top') if title else None
		self.values = values
		self.extCommand = command
		self.number = values[0]

		#Receives an index from the slider and sets the value
		def setText(idx):
			self.number = self.values[int(idx)]
			self.text.configure(text=self.values[int(idx)])
			if self.extCommand is not None: self.extCommand(self.values[int(idx)])

		self.slide = tk.Scale(self, command=setText,
			bg=bg, showvalue=0, from_=0, to=len(self.values)-1, font=font, **kwargs
		)
		self.text = tk.Label(self, font=font, width=4, bg=bg)
		self.slide.pack(side='right', expand=1, fill='x')
		self.text.pack(side='top', fill='both', padx=5)

	def get(self): return self.number
	def set(self, val):
		"""Receive a value and set the slider to the corresponding index."""
		self.number = val
		if val in self.values: self.slide.set(self.values.index(val))
		self.text.configure(text=val)

	def enable(self): self.disabled(False)
	def disable(self): self.disabled(True)
	def disabled(self, val):
		if val:
			self.text.configure(fg='#999')
			if self.label is not None: self.label.configure(fg='#999')
			self.slide.configure(state='disabled')
		else:
			self.text.configure(fg='#333')
			if self.label is not None: self.label.configure(fg='#333')
			self.slide.configure(state='normal')

	#Here for compatibility with other Tkinter widgets
	def configure(self, state): self.disabled(state=='disabled')

class expandableFrame(tk.Frame):
	"""A frame that can be expanded and collapsed by clicking on the title."""
	def __init__(self, parent=None, text: str='', fg='#333', bg='#FFF', padx: int=8, pady=None, font=None, startOpen=True):
		tk.Frame.__init__(self, parent, bg=bg)
		self.columnconfigure(1, weight=1)

		self.pady=pady
		self.padx=padx

		self.text = text
		self.titleLabel = tk.Label(self, fg=fg, bg=bg, font=font, cursor='hand2')
		self.titleLabel.bind('<Button-1>', self.toggle)
		self.titleLabel.grid(row=0, column=1, sticky='we', pady=(pady, 2))

		#Possibility of adding buttons on either side
		class btn(tk.Label):
			def __init__(self2, *args, **kwargs):
				self2.active = False
				self2.command = None
				hoverbg = Color(bg).darken().hex
				normalfg = Color(fg).lighten().hex
				hoverfg = Color(fg).lighten(1).hex
				super().__init__(*args, width=2, bg=bg, fg=normalfg, **kwargs)

				def hover(event):
					if self2.active: self2.config(bg=hoverbg, fg=hoverfg)
				def leave(event):
					if self2.active: self2.config(bg=bg, fg=normalfg)

				self2.bind('<Button-1>', self2.execute, add='+')
				self2.bind('<Enter>', hover, add='+')
				self2.bind('<Leave>', leave, add='+')

			def setup(self2, label, command, desc=None):
				self2.command = command
				self2.active = True
				self2.config(text=label)
				if desc: Tooltip(self2, desc)

			def execute(self2, e=None):
				if self2.active and self2.command is not None: self2.command()

		self.buttons = { 'left': btn(self), 'right': btn(self) }
		self.buttons['left'].grid(row=0, column=0, sticky='w', padx=5)
		self.buttons['right'].grid(row=0, column=2, sticky='e', padx=5)

		self.subframe = tk.Frame(self, padx=padx, pady=0, bg=bg)
		self._open = tk.IntVar()
		self.open = startOpen

	def toggle(self, event=None): self.open = not self.open

	@property
	def open(self): return bool(self._open.get())

	@open.setter
	def open(self, val):
		if val: #open
			self.subframe.grid(row=1, column=0, columnspan=3, padx=self.padx, pady=0, sticky='we')
			self.titleLabel['text'] = self.text+' '+'▾'
			self._open.set(1)
		else: #close
			self.subframe.grid_forget()
			self.titleLabel['text'] = self.text+' '+'▸'
			self._open.set(0)

class textCheck(tk.Label):
	"""A checkbox-like widget whose toggle is the entire element. `bg` and `fg` take a two-element color tuple for inactive and active states. https://helipad.dev/functions/textcheck/"""
	def __init__(self, parent=None, text=None, bg=('#FFFFFF','#419BF9'), fg=('#333333','#FFFFFF'),
		defaultValue=False, anchor='w', desc=None, callback=None
	):
		super().__init__(parent, text=text, bg=bg[defaultValue], fg=fg[defaultValue], anchor=anchor)
		self.bg = (Color(bg[0]), Color(bg[1]))
		self.fg = (Color(fg[0]), Color(fg[1]))

		#Generate disabled and hover colors
		self.disabledbg = (self.bg[0], self.bg[1].lighten(2))
		self.disabledfg = (self.fg[0].lighten(2), self.fg[1].lighten(2))
		hoverbg = (self.bg[0].darken(), self.bg[1].darken())
		hoverfg = (self.fg[0].darken(), self.fg[1].darken())

		self.value = defaultValue
		self.enabled = True

		def hover(event):
			if self.enabled: self.config(bg=hoverbg[self.value].hex, fg=hoverfg[self.value].hex)
		def leave(event):
			if self.enabled: self.config(bg=self.bg[self.value].hex, fg=self.fg[self.value].hex)

		if desc: Tooltip(self, desc)
		self.bind('<Button-1>', self.toggle, add='+')
		self.bind('<Enter>', hover, add='+')
		self.bind('<Leave>', leave, add='+')
		self.callback = callback

	def get(self): return self.value
	def set(self, value):
		if not self.enabled: return

		self.value = bool(value)
		self.config(bg=self.bg[self.value].hex, fg=self.fg[self.value].hex)
		if self.callback is not None: self.callback(value)

	def toggle(self, event=None): self.set(not self.value)

	def enable(self): self.disabled(False)
	def disable(self): self.disabled(True)
	def disabled(self, disable):
		if disable:
			bg = self.disabledbg
			fg = self.disabledfg
		else:
			bg = self.bg
			fg = self.fg

		self.enabled = not bool(disable)
		self.config(bg=bg[self.value].hex, fg=fg[self.value].hex)

class checkEntry(tk.Frame):
	"""A checkbox that enables/disables a text box. https://helipad.dev/functions/checkentry/"""
	def __init__(self, parent=None, title=None, width=20, bg='#FFFFFF', padx=0, pady=0, default='', datatype='string', command=None, limits=(0, 10**100)):
		tk.Frame.__init__(self, parent, bg=bg, padx=padx, pady=pady)
		self.callback = command

		#If we're enforcing int, don't allow nonnumerical input
		self.datatype=datatype
		def validate(code, insert, oldval, newval):
			if not self.enabled: return
			allow = True
			if self.datatype=='int' and code == '1':
				for c in insert:
					if c not in '0123456789':
						allow = False
			if allow: self.callback(int(newval) if self.datatype=='int' and newval!='' else newval)
			return allow
		valint = self.register(validate)

		self.enabled = True
		self.entryValue = tk.StringVar()
		self.entryValue.set(default)
		if datatype=='string':
			self.textbox = tk.Entry(self, textvariable=self.entryValue, width=width, state='disabled', validate='key', validatecommand=(valint, '%d', '%S', '%s', '%P'), highlightbackground=bg)
		elif datatype=='int':
			self.textbox = tk.Spinbox(self, from_=limits[0], to=limits[1], textvariable=self.entryValue, width=width, state='disabled', validate='key', validatecommand=(valint, '%d', '%S', '%s', '%P'), highlightbackground=bg)
		else: raise ValueError(ï('Invalid Checkentry datatype. Must be either "string" or "int"'))
		self.textbox.grid(row=0, column=1)

		self.checkVar = tk.BooleanVar()
		self.checkbox = tk.Checkbutton(self, text=title, bg=bg, var=self.checkVar, onvalue=True, offvalue=False, command=self.disableTextfield)
		self.checkbox.grid(row=0, column=0)

	def disableTextfield(self):
		"""Update the enabled state of the text field to correspond to the value of the checkbox. https://helipad.dev/functions/checkentry/disabletextfield/"""
		self.textbox.config(state='normal' if self.checkVar.get() else 'disabled')
		if callable(self.callback): self.callback(self.get())

	def get(self):
		"""Retrieve the value of the combined input. Returns `False` if the checkbox is unchecked, and the value of the textbox otherwise. https://helipad.dev/functions/checkentry/get/"""
		if not self.checkVar.get(): return False
		v = self.entryValue.get()
		if self.datatype=='int':
			return 0 if v=='' else int(v)
		else: return v

	def set(self, val):
		"""Set the input value. Can receive a `bool` to toggle the checkbox, or a `string` or an `int` (depending on the datatype) to change the value of the textbox. https://helipad.dev/functions/checkentry/set/"""
		if isinstance(val, bool):
			self.checkVar.set(val)
		elif isinstance(val, (str, int)):
			if self.datatype=='int' and val!='': val=int(val)
			self.checkVar.set(True)
			self.entryValue.set(val)
		self.disableTextfield()

	def enable(self):
		"""Enable the `checkEntry` element so it can be interacted with by the user. https://helipad.dev/functions/checkentry/enable/"""
		self.disabled(False)
	def disable(self):
		"""Disable the `checkEntry` element so it cannot be interacted with by the user.https://helipad.dev/functions/checkentry/disable/"""
		self.disabled(True)
	def disabled(self, disable):
		"""Enables or disables the `checkEntry` element. https://helipad.dev/functions/checkentry/disabled/"""
		self.checkbox.config(state='disabled' if disable else 'normal')
		self.textbox.config(state='disabled' if disable or not self.checkVar.get() else 'normal')
		self.enabled = not disable

	#Here for compatibility with other Tkinter widgets
	def configure(self, state): self.disabled(state=='disabled')

class checkGrid(expandableFrame):
	"""An `expandableFrame` full of `textCheck`s, with setters and getters. https://helipad.dev/functions/checkentry/"""
	def __init__(self, parent=None, text: str='', columns: int=3, fg='#333', bg='#FFF', padx: int=8, pady: int=5, startOpen=True, callback=None):
		super().__init__(parent=parent, text=text, fg=fg, bg=bg, padx=padx, pady=pady, font=Cpanel.font, startOpen=startOpen)
		self.bg = bg
		self.columns = columns
		self.checks = {}
		self._index=0
		self.callback = callback

		for i in range(columns): self.subframe.columnconfigure(i, weight=1)
		self.buttons['right'].setup('✔︎', self.toggleAll)

	def addCheck(self, var, text: str, defaultValue: bool=True, desc: bool=None):
		"""Adds a `textCheck` element to the frame. https://helipad.dev/functions/checkgrid/addcheck/"""
		if self.callback is not None:
			def cbWrap(val): self.callback((var, val))
		else: cbWrap = None

		self.checks[var] = textCheck(self.subframe, text=text, anchor='w', defaultValue=defaultValue, bg=(self.bg, '#419BF9'), desc=desc, callback=cbWrap)
		self.checks[var].grid(row=int(ceil(len(self.checks)/self.columns)), column=(len(self.checks)-1)%self.columns, sticky='we')
		return self.checks[var]

	def toggleAll(self):
		if [True for c in self.checks.values() if c.get()]:
			for c in self.checks.values(): c.set(False)
		else:
			for c in self.checks.values(): c.set(True)

	def __getitem__(self, index): return self.checks[index].get()
	def __setitem__(self, index, value): self.checks[index].set(value)

	def items(self): return self.checks.items()
	def values(self): return self.checks.values()
	def keys(self): return self.checks.keys()
	def disabled(self, key, val=None):
		if isinstance(key, bool):
			if key: self.disable()
			else: self.enable()
		else: self.checks[key].disabled(val)
	def enable(self, key=None):
		if key: self.checks[key].enable()
		else:
			for c in self.values(): c.enable()
	def disable(self, key=None):
		if key: self.checks[key].disable()
		else:
			for c in self.values(): c.disable()
	def configure(self, state): self.disabled(state=='disabled')

class Tooltip:
	"""Replaces `PMW.balloon` because it's unreliably maintained. Modified from https://stackoverflow.com/questions/3221956/how-do-i-display-tooltips-in-tkinter"""
	def __init__(self, widget, text, bg='#FFFFEA', pad=(5, 3, 5, 3), waittime=500, wraplength=250, tip_delta=(10, 5)):
		for v in ['text', 'widget', 'bg', 'pad', 'waittime', 'wraplength', 'tip_delta']: setattr(self, v, locals()[v]) #Populate object with arguments
		self.widget.bind("<Enter>", self.onEnter)
		self.widget.bind("<Leave>", self.onLeave)
		self.widget.bind("<ButtonPress>", self.onLeave)
		self.id = None
		self.tw = None

	def onLeave(self, event=None):
		self.unschedule()
		self.hide()

	def onEnter(self, event=None):
		self.unschedule()
		self.id = self.widget.after(self.waittime, self.show)

	def unschedule(self):
		id_ = self.id
		self.id = None
		if id_: self.widget.after_cancel(id_)

	def show(self):
		self.tw = tk.Toplevel(self.widget) # creates a toplevel window
		self.tw.wm_overrideredirect(True) # Leaves only the label and removes the app window

		win = tk.Frame(self.tw, background=self.bg, borderwidth=0)
		label = tk.Label(win, text=self.text, justify='left', background=self.bg, relief='solid', borderwidth=0, wraplength=self.wraplength)
		label.grid(padx=(self.pad[0], self.pad[2]), pady=(self.pad[1], self.pad[3]), sticky='nsew')
		win.grid()

		#Calculate tooltip position
		s_width, s_height = self.widget.winfo_screenwidth(), self.widget.winfo_screenheight()
		width, height = (self.pad[0] + label.winfo_reqwidth() + self.pad[2], self.pad[1] + label.winfo_reqheight() + self.pad[3])
		mouse_x, mouse_y = self.widget.winfo_pointerxy()
		x1, y1 = mouse_x + self.tip_delta[0], mouse_y + self.tip_delta[1]
		x2, y2 = x1 + width, y1 + height

		x_delta = max(x2 - s_width, 0)
		y_delta = max(y2 - s_height, 0)

		#If offscreen
		if (x_delta, y_delta) != (0, 0):
			if x_delta: x1 = mouse_x - self.tip_delta[0] - width
			if y_delta: y1 = mouse_y - self.tip_delta[1] - height
		y1 = max(y1, 0)

		self.tw.wm_geometry(f'+{x1}+{y1}')

	def hide(self):
		if self.tw: self.tw.destroy()
		self.tw = None


#Registers parameters of every available type in order to test the rendering of the control panel, and runs a dummy model to test the Charts visualizer.

#===============
# PREAMBLE
#===============

from random import uniform
from helipad import Helipad
heli = Helipad()
heli.name = 'Test'

#A handful of breeds and goods
breeds = [
	('hobbit', 'jam', '#D73229'),
	('dwarf', 'axe', '#2D8DBE'),
	('elf', 'lembas', '#CCBB22')
]
AgentGoods = {}
for b in breeds:
	heli.agents.addBreed(b[0], b[2], prim='agent')
	heli.goods.add(b[1], b[2])

def gcallback(model, name, val):
	print(name, '=', val)

def icallback(model, name, item, val):
	print(name, '/', item, '=', val)

#===============
# ADD PARAMETERS
#===============

heli.params.add('gslider', 'Global slider', 'slider', dflt=1.5, opts={'low': 1, 'high': 5, 'step': 0.1}, callback=gcallback, desc="A global slider")
heli.params.add('gcheck', 'Global check', 'check', dflt=True, callback=gcallback, desc="A global checkbox")
heli.params.add('gmenu', 'Global menu', 'menu', dflt='two', opts={
	'one': 'Option one',
	'two': 'Option two',
	'three': 'Option three'
}, callback=gcallback, desc="A global menu")
heli.params.add('gcheckentry', 'Global Checkentry', 'checkentry', dflt='They\'re taking the hobbits to Isengard', callback=gcallback, desc="A global checkentry")
heli.params.add('glogslider', 'Global Logslider', 'slider', dflt=8, opts=[1,2,3,5,8,13,21,34], callback=gcallback, desc="A global logslider")

heli.params.add('islider', 'Item Slider', 'slider', per='breed', dflt={'hobbit': 0.1, 'dwarf': 0.3}, opts={'low':0, 'high': 1, 'step': 0.01}, callback=icallback, desc="A per-breed slider")
heli.params.add('icheck', 'Item Check', 'check', per='good', dflt={'jam': False, 'axe': True}, callback=icallback, desc="A per-good slider")
heli.params.add('imenu', 'Item Menu', 'menu', per='breed', dflt={'hobbit': 'three', 'dwarf': 'two'}, opts={
	'one': 'Option one',
	'two': 'Option two',
	'three': 'Option three'
}, callback=icallback, desc="A per-breed menu")
heli.params.add('icheckentry', 'Item Checkentry', 'checkentry', per='good', dflt={'jam': False, 'axe': 'wood'}, callback=icallback, desc="A per-good checkentry")
# heli.params.add('ilogslider', 'Item Logslider', 'slider', per='good', dflt={'axe': 5, 'lembas': 21}, opts=[1,2,3,5,8,13,21,34], callback=icallback, desc="A per-item logslider")

heli.params.add('gcheckgrid', 'Global Checkgrid', 'checkgrid',
	opts={'gondor':('Gondor', 'Currently calling for aid'), 'isengard':'Isengard', 'rohan':'Rohan', 'rivendell':'Rivendell', 'khazad':('Khazad-dûm', 'Nice diacritic')},
	dflt=['gondor', 'rohan', 'khazad'], callback=gcallback, desc="A global checkgrid"
)
heli.param('num_agent', 18)

#===============
# A DUMMY MODEL
# Testing the bar chart and network visualizers
#===============

import random
from helipad.visualize import Charts
viz = heli.useVisual(Charts)

@heli.hook
def agentInit(agent, model):
	for i in range(20): setattr(agent, 'prop'+str(i), 0)
	agent.s1 = 0
	agent.s2 = 0

@heli.hook
def agentStep(agent, model, stage):
	for i in range(20):
		v = getattr(agent, 'prop'+str(i))
		setattr(agent, 'prop'+str(i), v-1 if random.randint(0, 1) else v+1)
	
	agent.s1 += uniform(0,agent.id+model.t**0.75)
	agent.s2 += uniform(0,agent.id+model.t**0.75)

@heli.hook
def modelPostSetup(model):
	model.agents.createNetwork(0.2)

def newedge(model):
	a1, a2 = random.choice(model.agents['agent']), random.choice(model.agents['agent'])
	while a1.edges.With(a2): a1, a2 = random.choice(model.agents['agent']), random.choice(model.agents['agent'])
	a1.edges.add(a2, direction=random.choice([True, False]), weight=random.choice([0.5,1,2,3]))

#Cut one edge and create one edge
@heli.hook
def modelPostStep(model):
	random.choice(model.agents.edges['edge']).cut()
	newedge(model)

net = viz.addPlot('net', 'Network Structure', type='agents') #scatter=['s1', 's2']
bar1 = viz.addPlot('prop', 'Bar Chart')
bar2 = viz.addPlot('prop2', 'Horizontal Bar Chart', horizontal=True)
net.config({
	'agentLabel': True,
 	'labelColor': '#FFFFFF',
# 	'labelFamily': 'serif',
#	'agentMarker': 'H',
	'labelAlpha': 0.75,
	'labelWeight': 'bold',
 	'labelSize': 8,
	'regLine': True
})

gcolors = ['F00', 'F03', 'F06', 'F09', 'F0C', 'C0F', '90F', '60F', '30F', '00F']
for i in range(20):
	heli.data.addReporter('prop'+str(i), heli.data.agentReporter('prop'+str(i), std=0.1))
	(bar1 if i<10 else bar2).addBar('prop'+str(i), str(i), '#'+gcolors[i%10])

#Replace an agent, but only if we click during the current model time
@heli.hook
def agentClick(agents, plot, t):
	if t != heli.t: return
	
	for agent in agents:
		new = agent.reproduce()
		enum = len(agent.edges[plot.network]) if plot.network in agent.edges else 0
		agent.die()
		for e in range(enum): newedge(heli)
		print('Killing agent',agent.id,'and creating agent',new.id)
	plot.update(None, t)
	plot.draw(t, forceUpdate=True)

#===============
# LAUNCH
#===============

heli.launchCpanel()


# A sample spatial model with agents eating grass off patches.

#===============
# SETUP
#===============

from random import choice, randint
from numpy import mean
from helipad import Helipad

heli = Helipad()
heli.name = 'Grass Eating'
heli.agents.order = 'random'
heli.stages = 5

heli.params.add('energy', 'Energy from grass', 'slider', dflt=2, opts={'low': 2, 'high': 10, 'step': 1})
heli.params.add('smart', 'Smart consumption', 'check', dflt=True, desc='Move to the neighboring patch with the most grass, rather than randomly')
heli.params.add('e2reproduce', 'Energy to reproduce', 'slider', dflt=25, opts={'low': 0, 'high': 100, 'step': 5})
heli.params.add('maleportion', 'Male portion reproduction', 'slider', dflt=40, opts={'low': 0, 'high': 100, 'step': 5})
heli.params.add('maxLife', 'Max Lifespan', 'slider', dflt=200, opts={'low': 100, 'high': 1000, 'step': 10})
heli.params.add('grassrate', 'Grass Rate', 'slider', dflt=10, opts={'low': 1, 'high': 100, 'step': 1})

heli.params['num_agent'].opts = {'low': 1, 'high': 200, 'step': 1}
heli.param('num_agent', 200)

heli.agents.addBreed('male', 'blue')
heli.agents.addBreed('female', 'pink')
heli.goods.add('energy', 'red', 5)

#===============
# BEHAVIOR
#===============

#Dividing it into stages like this (like in the NetLogo version) appears to make it more viable,
#perhaps because it encourages bunching onto the same fecund patch, with more opportunities for
#reproduction, whereas if they do it all at once, they avoid each other too much

@heli.hook
def agentStep(agent, model, stage):

	#Look for the neighboring patch with the most grass and move to it, if smart
	if stage==1:
		if model.param('smart'):
			maxenergy = max(n.stocks['energy'] for n in agent.patch.neighbors)
			prospects = [n for n in agent.patch.neighbors if n.stocks['energy'] == maxenergy]
		else: prospects = agent.patch.neighbors
		agent.orientTo(choice(prospects))
		agent.forward()
		agent.stocks['energy'] -= 1

	#Eat grass
	elif stage==2:
		if agent.patch.stocks['energy'] > 0:
			agent.patch.stocks['energy'] -= 1
			agent.stocks['energy'] += model.param('energy')

	#Reproduce
	elif stage==3:
		if agent.breed == 'male':
			e = model.param('e2reproduce')
			p = model.param('maleportion')
			me, fe = e*p/100, e*(100-p)/100
			if agent.stocks['energy'] > me:
				prospects = [f for f in agent.patch.agentsOn if f.breed=='female' and f.stocks['energy']>fe]
				if len(prospects):
					mate = choice(prospects)
					agent.stocks['energy'] -= me
					mate.stocks['energy'] -= fe
					child = mate.reproduce(inherit=[('breed', 'rand')], partners=[agent])
					child.stocks['energy'] = e

	#Die
	elif stage==4:
		if agent.stocks['energy'] <= 0 or agent.age > model.param('maxLife'):
			agent.die()
			model.deathAge.append(agent.age)
			if len(model.deathAge) > 100: model.deathAge.pop(0)

@heli.hook
def patchStep(patch, model, stage):
	#Regrow grass
	if stage==5 and patch.stocks['energy'] < 5 and randint(1,100) <= model.param('grassrate'):
		patch.stocks['energy'] += 1

@heli.hook
def modelPostSetup(model):
	model.deathAge = []

#Stop the model when we have no more females left to reproduce
@heli.event
def nofemales(model): return len(model.agents['agent']['female']) <= 1
heli.param('stopafter', 'nofemales')
heli.param('refresh', 1)

#===============
# DATA AND VISUALIZATION
#===============

heli.data.addReporter('grass', heli.data.agentReporter('stocks', 'patch', good='energy', stat='sum'))
heli.data.addReporter('age', heli.data.agentReporter('age', 'agent'))
heli.data.addReporter('num_agent', lambda model: len(model.agents['agent']))
heli.data.addReporter('sexratio', lambda model: len(model.agents['agent']['male'])/len(model.agents['agent']['female']))
heli.data.addReporter('expectancy', lambda model: mean(model.deathAge))
heli.data.addReporter('agentenergy', heli.data.agentReporter('stocks', 'agent', good='energy', percentiles=[0,100]))

mapPlot = heli.spatial(dim=16, corners=True)
mapPlot.scatter = ['age', 'good:energy']
mapPlot.config({
	'patchProperty': 'good:energy',
	'patchColormap': 'Greens',
	'agentSize': 'good:energy'
})

# # Only forward 0.368
# import geopandas,os
# __location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
# gdf = geopandas.read_file(os.path.join(__location__, 'NC_Counties.geojson'))
# for i,county in gdf.iterrows():
#     heli.patches.add(county['geometry'], county['CO_NAME'])

pop = heli.visual.addPlot('pop', 'Population', 'timeseries', logscale=True, selected=False)
sexratio = heli.visual.addPlot('sexratio', 'Sex Ratio', 'timeseries', logscale=True, selected=False)
age = heli.visual.addPlot('age', 'Age', 'timeseries', selected=False)
energy = heli.visual.addPlot('energy', 'Energy', 'timeseries', selected=False)

pop.addSeries('num_agent', 'Population', 'black')
pop.addSeries('grass', 'Grass', 'green')
sexratio.addSeries('sexratio', 'M/F Sex Ratio', 'brown')
age.addSeries('age', 'Average Age', 'blue')
pop.addSeries('expectancy', 'Life Expectancy', 'black')
energy.addSeries('agentenergy', 'Energy', 'green')

@heli.hook
def agentClick(agent, plot, t):
	print([f'Agent {a.id} at ({a.x}, {a.y})' for a in agent if a is not None])

@heli.hook
def patchClick(patch, plot, t):
	print(patch)

#===============
# LAUNCH THE GUI
#===============

heli.launchCpanel()